from typing import TYPE_CHECKING

from ..adapter.model import Event
from ..ctx import LoggerCtx
from ..log.base import LogLevel
from ..utils import RWContext
from .process import Flow

if TYPE_CHECKING:
    from ..plugin.base import Plugin


class EventHandler:
    def __init__(self, plugin: "Plugin", flow: Flow) -> None:
        self.flow = flow
        self.logger = LoggerCtx().get()
        self.name = flow.name

        self._plugin = plugin
        self._handle_ctrl = RWContext()
        self._temp = flow.temp
        self.invalid = False

    async def _handle_event(self, event: Event) -> None:
        try:
            await self.flow.run(event)
        except Exception:
            self.logger.exception(f"事件处理流 {self.name} 发生异常")
            self.logger.generic_obj(
                f"异常点 event {event.id}", event.__dict__, level=LogLevel.ERROR
            )
            self.logger.generic_obj("异常点局部变量：", locals(), level=LogLevel.ERROR)

    async def handle(self, event: Event) -> None:
        if self.invalid:
            return

        if not self._temp:
            async with self._handle_ctrl.read():
                if self.invalid:
                    return
                return await self._handle_event(event)

        async with self._handle_ctrl.write():
            if self.invalid:
                return
            await self._handle_event(event)
            self.invalid = True
            return

    async def expire(self) -> None:
        async with self._handle_ctrl.write():
            self.invalid = True



from ._meta import MetaInfo, __version__
from .adapter import Action, Adapter, Echo, Event
from .adapter.generic import send_image, send_text
from .bot import Bot, get_bot
from .ctx import Context
from .di import Depends
from .handle import Flow, FlowStore, node, rewind, stop
from .log import GenericLogger, Logger, LogLevel, get_logger
from .plugin import AsyncShare, Plugin, PluginLifeSpan, SyncShare
from .session import Rule, SessionStore, enter_session, suspend
from .typ import HandleLevel, LogicMode



from __future__ import annotations

from asyncio import create_task
from dataclasses import dataclass
from itertools import tee
from typing import Iterable, NoReturn, Sequence

from ..adapter.model import Event
from ..ctx import FlowCtx, FlowRecord, FlowRecords
from ..ctx import FlowRecordStage as RecordStage
from ..ctx import FlowStatus, FlowStore
from ..di import DependNotMatched, inject_deps
from ..exceptions import FlowError
from ..typ import AsyncCallable, HandleLevel
from ..utils import get_obj_name

_FLOW_CTX = FlowCtx()


def node(func: AsyncCallable[..., bool | None]) -> FlowNode:
    """处理结点装饰器，将当前异步可调用对象装饰为一个处理结点"""
    return FlowNode(func)


def no_deps_node(func: AsyncCallable[..., bool | None]) -> FlowNode:
    """与 :func:`node` 类似，但是不自动为结点标记依赖注入。

    需要后续使用 :func:`.inject_deps` 手动标记依赖注入，
    这适用于某些对处理结点进行再装饰的情况
    """
    return FlowNode(func, no_deps=True)


class FlowNode:
    """处理流结点"""

    def __init__(
        self, func: AsyncCallable[..., bool | None], no_deps: bool = False
    ) -> None:
        self.name = get_obj_name(func, otype="callable")
        self.processor: AsyncCallable[..., bool | None] = (
            func if no_deps else inject_deps(func)
        )

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(name={self.name})"

    async def process(self, event: Event, flow: Flow) -> None:
        try:
            status = _FLOW_CTX.get()
            records, store = status.records, status.store
        except _FLOW_CTX.lookup_exc_cls:
            records, store = FlowRecords(), FlowStore()

        with _FLOW_CTX.unfold(FlowStatus(event, flow, self, True, records, store)):
            try:
                records.append(
                    FlowRecord(RecordStage.NODE_START, flow.name, self.name, event)
                )

                try:
                    ret = await self.processor()
                    records.append(
                        FlowRecord(RecordStage.NODE_FINISH, flow.name, self.name, event)
                    )
                except DependNotMatched as e:
                    ret = False
                    records.append(
                        FlowRecord(
                            RecordStage.DEPENDS_NOT_MATCH,
                            flow.name,
                            self.name,
                            event,
                            f"Real({e.real_type}) <=> Annotation({e.hint})",
                        )
                    )

                if ret in (None, True) and _FLOW_CTX.get().next_valid:
                    await nextn()

            except FlowContinued:
                await nextn()


@dataclass
class _NodeInfo:
    nexts: list[FlowNode]
    in_deg: int
    out_deg: int

    def copy(self) -> _NodeInfo:
        return _NodeInfo(self.nexts, self.in_deg, self.out_deg)


class Flow:
    """处理流

    :ivar str name: 处理流的标识
    """

    def __init__(
        self,
        name: str,
        *edge_maps: Iterable[Iterable[FlowNode] | FlowNode],
        priority: HandleLevel = HandleLevel.NORMAL,
        temp: bool = False,
    ) -> None:
        """初始化处理流

        :param name: 处理流的标识
        :param edge_maps: 边映射，遵循 melobot 的 graph edges 表示方法
        :param priority: 处理流的优先级
        :param temp: 处理流是否只运行一次
        """
        self.name = name
        self.temp = temp
        self.graph: dict[FlowNode, _NodeInfo] = {}

        self._priority = priority
        self._priority_cb: AsyncCallable[[HandleLevel], None]

        _edge_maps = tuple(
            tuple((elem,) if isinstance(elem, FlowNode) else elem for elem in emap)
            for emap in edge_maps
        )
        edges = self._get_edges(_edge_maps)

        for n1, n2 in edges:
            self._add(n1, n2)

        if not self._valid_check():
            raise FlowError(f"定义的处理流 {self.name} 中存在环路")

    def _get_edges(
        self, edge_maps: Sequence[Sequence[Iterable[FlowNode]]]
    ) -> list[tuple[FlowNode, FlowNode]]:
        edges: list[tuple[FlowNode, FlowNode]] = []

        for emap in edge_maps:
            iter1, iter2 = tee(emap, 2)
            try:
                next(iter2)
            except StopIteration:
                continue

            if len(emap) == 1:
                for n in emap[0]:
                    self._add(n, None)
                continue

            for from_seq, to_seq in zip(iter1, iter2):
                for n1 in from_seq:
                    for n2 in to_seq:
                        if (n1, n2) not in edges:
                            edges.append((n1, n2))

        return edges

    @property
    def priority(self) -> HandleLevel:
        """处理流的优先级"""
        return self._priority

    def on_priority_reset(self, callback: AsyncCallable[[HandleLevel], None]) -> None:
        self._priority_cb = callback

    async def reset_priority(self, new_prior: HandleLevel) -> None:
        """重设处理流的优先级

        不会立即生效，通常会在下一次进入处理流前生效。
        因此返回时不代表已经更改优先级，只是添加了计划重设优先级的任务

        :param new_prior: 新优先级
        """

        async def _reset_wrapper() -> None:
            await self._priority_cb(new_prior)
            self._priority = new_prior

        create_task(_reset_wrapper())

    @property
    def starts(self) -> tuple[FlowNode, ...]:
        return tuple(n for n, info in self.graph.items() if info.in_deg == 0)

    @property
    def ends(self) -> tuple[FlowNode, ...]:
        return tuple(n for n, info in self.graph.items() if info.out_deg == 0)

    def __repr__(self) -> str:
        output = f"{self.__class__.__name__}(name={self.name}, nums={len(self.graph)}"

        if len(self.graph):
            output += f", starts=[{', '.join(repr(n) for n in self.starts)}])"
        else:
            output += ")"
        return output

    def _add(self, _from: FlowNode, to: FlowNode | None) -> None:
        from_info = self.graph.setdefault(_from, _NodeInfo([], 0, 0))

        if to is not None:
            to_info = self.graph.setdefault(to, _NodeInfo([], 0, 0))
            to_info.in_deg += 1
            from_info.out_deg += 1
            from_info.nexts.append(to)

    def _valid_check(self) -> bool:
        graph = {n: info.copy() for n, info in self.graph.items()}

        while len(graph):
            for n, info in graph.items():
                nexts, in_deg = info.nexts, info.in_deg

                if in_deg == 0:
                    graph.pop(n)
                    for next_n in nexts:
                        graph[next_n].in_deg -= 1
                    break

            else:
                return False

        return True

    def link(self, flow: Flow, priority: HandleLevel | None = None) -> Flow:
        """连接另一处理流返回新处理流，并设置新优先级

        :param flow: 连接的新流
        :param priority: 新优先级，若为空，则使用两者中较小的优先级
        :return: 新的处理流
        """
        _froms = self.ends
        _tos = flow.starts
        new_edges = tuple((n1, n2) for n1 in _froms for n2 in _tos)

        new_flow = Flow(
            f"{self.name} ~ {flow.name}",
            *new_edges,
            priority=priority if priority else min(self.priority, flow.priority),
            temp=self.temp or flow.temp,
        )

        for n1, info in (self.graph | flow.graph).items():
            if not len(info.nexts):
                new_flow._add(n1, None)
                continue
            for n2 in info.nexts:
                new_flow._add(n1, n2)

        if not self._valid_check():
            raise FlowError(f"定义的处理流 {self.name} 中存在环路")

        return new_flow

    async def run(self, event: Event) -> None:
        if not len(self.starts):
            return

        try:
            status = _FLOW_CTX.get()
            records, store = status.records, status.store
        except _FLOW_CTX.lookup_exc_cls:
            records, store = FlowRecords(), FlowStore()

        with _FLOW_CTX.unfold(
            FlowStatus(event, self, self.starts[0], True, records, store)
        ):
            try:
                records.append(
                    FlowRecord(
                        RecordStage.FLOW_START, self.name, self.starts[0].name, event
                    )
                )

                idx = 0
                while idx < len(self.starts):
                    try:
                        await self.starts[idx].process(event, self)
                        idx += 1
                    except FlowRewound:
                        pass

                records.append(
                    FlowRecord(
                        RecordStage.FLOW_FINISH, self.name, self.starts[0].name, event
                    )
                )
            except FlowBroke:
                pass


class _FlowSignal(BaseException): ...


class FlowBroke(_FlowSignal): ...


class FlowContinued(_FlowSignal): ...


class FlowRewound(_FlowSignal): ...


async def nextn() -> None:
    """运行下一处理结点（在处理结点中使用）"""
    try:
        status = _FLOW_CTX.get()
        nexts = status.flow.graph[status.node].nexts
        if not status.next_valid:
            return

        idx = 0
        while idx < len(nexts):
            try:
                await nexts[idx].process(status.event, status.flow)
                idx += 1
            except FlowRewound:
                pass

    except _FLOW_CTX.lookup_exc_cls:
        raise FlowError("此时不在活动的事件处理流中，无法调用下一处理结点") from None
    finally:
        status.next_valid = False


async def block() -> None:
    """阻止当前事件向更低优先级的处理流传播（在处理结点中使用）"""
    status = _FLOW_CTX.get()
    status.records.append(
        FlowRecord(RecordStage.BLOCK, status.flow.name, status.node.name, status.event)
    )
    status.event.spread = False


async def stop() -> NoReturn:
    """立即停止当前处理流（在处理结点中使用）"""
    status = _FLOW_CTX.get()
    status.records.append(
        FlowRecord(RecordStage.STOP, status.flow.name, status.node.name, status.event)
    )
    status.records.append(
        FlowRecord(
            RecordStage.NODE_EARLY_FINISH,
            status.flow.name,
            status.node.name,
            status.event,
        )
    )
    status.records.append(
        FlowRecord(
            RecordStage.FLOW_EARLY_FINISH,
            status.flow.name,
            status.node.name,
            status.event,
        )
    )
    raise FlowBroke("事件处理流被安全地提早结束，请无视这个内部工作信号")


async def bypass() -> NoReturn:
    """立即跳过当前处理结点剩下的步骤，运行下一处理结点（在处理结点中使用）"""
    status = _FLOW_CTX.get()
    status.records.append(
        FlowRecord(RecordStage.BYPASS, status.flow.name, status.node.name, status.event)
    )
    status.records.append(
        FlowRecord(
            RecordStage.NODE_EARLY_FINISH,
            status.flow.name,
            status.node.name,
            status.event,
        )
    )
    raise FlowContinued("事件处理流安全地跳过结点执行，请无视这个内部工作信号")


async def rewind() -> NoReturn:
    """立即重新运行当前处理结点（在处理结点中使用）"""
    status = _FLOW_CTX.get()
    status.records.append(
        FlowRecord(RecordStage.REWIND, status.flow.name, status.node.name, status.event)
    )
    status.records.append(
        FlowRecord(
            RecordStage.NODE_EARLY_FINISH,
            status.flow.name,
            status.node.name,
            status.event,
        )
    )
    raise FlowRewound("事件处理流安全地重复执行处理结点，请无视这个内部工作信号")


async def flow_to(flow: Flow) -> None:
    """立即进入一个其他处理流（在处理结点中使用）"""
    status = _FLOW_CTX.get()
    await flow.run(status.event)



from .base import (
    AbstractEchoFactory,
    AbstractEventFactory,
    AbstractOutputFactory,
    Adapter,
    AdapterLifeSpan,
)
from .content import Content
from .model import Action, ActionChain, ActionHandle, Echo, Event, open_chain



import asyncio
from asyncio import Queue, create_task

from melobot.bot import Bot
from melobot.log import GenericLogger
from melobot.plugin import Plugin
from melobot.protocols.onebot.v11 import handle
from melobot.protocols.onebot.v11.adapter.base import Adapter
from melobot.protocols.onebot.v11.adapter.event import MessageEvent
from melobot.protocols.onebot.v11.io.base import BaseIO
from melobot.protocols.onebot.v11.io.packet import EchoPacket, InPacket, OutPacket
from melobot.protocols.onebot.v11.utils import (
    CmdParser,
    GroupMsgChecker,
    LevelRole,
    ParseArgs,
)
from tests.base import *

_GRUOP_EVENT_DICT = {
    "time": 1725292489,
    "self_id": 123456,
    "post_type": "message",
    "message_type": "group",
    "sub_type": "normal",
    "sender": {
        "age": 0,
        "nickname": "这是一个群昵称",
        "sex": "unknown",
        "user_id": 3,
        "area": "",
        "card": "",
        "level": "",
        "role": "member",
        "title": "",
    },
    "message_id": -1234567890,
    "font": 0,
    "message": "\n\r\t .echo/123/456    \r\n",
    "user_id": 3,
    "anonymous": None,
    "group_id": 6,
    "raw_message": "45691237\n\r\t .echo/123/456    \r\n",
}


h = handle.on_start_match(
    ["123", "456"],
    checker=GroupMsgChecker(
        role=LevelRole.WHITE,
        owner=1,
        super_users=[2, 3],
        white_users=[4],
        black_users=[5],
        white_groups=[6, 7],
    ),
    parser=CmdParser(".", "/", "echo"),
)


@h
async def test_this(
    bot: Bot,
    event: MessageEvent,
    logger: GenericLogger,
    args: ParseArgs = handle.GetParseArgs(),
) -> None:
    logger.info(args)
    await bot.close()
    _SUCCESS_SIGNAL.set()


class TempPlugin(Plugin):
    version = "1.0.0"
    flows = [test_this]


class TempIO(BaseIO):
    def __init__(self) -> None:
        super().__init__(1)
        self.queue = Queue()
        self.queue.put_nowait(InPacket(data=_GRUOP_EVENT_DICT))

    async def open(self) -> None:
        pass

    def opened(self) -> bool:
        return True

    async def close(self) -> None:
        pass

    async def input(self) -> InPacket:
        return await self.queue.get()

    async def output(self, packet: OutPacket) -> EchoPacket:
        if packet.echo_id is None:
            return EchoPacket(noecho=True)


_SUCCESS_SIGNAL = asyncio.Event()


async def test_adapter_base():
    mbot = Bot("test_handle")
    mbot.add_io(TempIO())
    mbot.add_adapter(Adapter())
    mbot.load_plugin(TempPlugin())
    create_task(mbot.core_run())
    await mbot._rip_signal.wait()
    await _SUCCESS_SIGNAL.wait()



from melobot.io import AbstractIOSource
from melobot.log import GenericLogger, get_logger
from melobot.typ import abstractmethod

from ..const import PROTOCOL_IDENTIFIER
from .packet import EchoPacket, InPacket, OutPacket


class BaseIO(AbstractIOSource[InPacket, OutPacket, EchoPacket]):
    """

    :ivar float cd_time: 发送行为操作的冷却时间（防风控）
    """

    # pylint: disable=duplicate-code
    def __init__(self, cd_time: float) -> None:
        super().__init__(PROTOCOL_IDENTIFIER)
        self.cd_time = cd_time

    @property
    def logger(self) -> GenericLogger:
        return get_logger()

    @abstractmethod
    async def open(self) -> None:
        raise NotImplementedError

    @abstractmethod
    async def close(self) -> None:
        raise NotImplementedError

    @abstractmethod
    def opened(self) -> bool:
        raise NotImplementedError

    @abstractmethod
    async def input(self) -> InPacket:
        raise NotImplementedError

    @abstractmethod
    async def output(self, packet: OutPacket) -> EchoPacket:
        raise NotImplementedError



import asyncio
import time
from enum import Enum
from typing import Any, Callable, Generic, TypeVar

from .ctx import LoggerCtx
from .di import inject_deps
from .log.base import LogLevel
from .typ import AsyncCallable, P

HookEnumT = TypeVar("HookEnumT", bound=Enum)


class HookRunner(Generic[HookEnumT]):
    def __init__(
        self, type: HookEnumT, func: AsyncCallable[..., None], once: bool = False
    ) -> None:
        self.type = type
        self.callback: AsyncCallable[..., None] = inject_deps(func, manual_arg=True)
        self.once = once
        self._valid = True
        self._lock = asyncio.Lock()

    async def _run(self, *args: Any, **kwargs: Any) -> None:
        try:
            await self.callback(*args, **kwargs)
        except Exception:
            logger = LoggerCtx().get()
            logger.exception(f"{self.type} 类型的 hook 方法 {self.callback} 发生异常")
            logger.generic_obj("异常点局部变量：", locals(), level=LogLevel.ERROR)

    async def run(self, *args: Any, **kwargs: Any) -> None:
        if not self._valid:
            return
        if not self.once:
            return await self._run(*args, **kwargs)

        async with self._lock:
            if not self._valid:
                return
            await self._run(*args, **kwargs)
            self._valid = False


class HookBus(Generic[HookEnumT]):
    def __init__(self, type: type[HookEnumT], tag: str | None = None) -> None:
        self._hooks: dict[HookEnumT, list[HookRunner]] = {t: [] for t in list(type)}
        self._stamps: dict[HookEnumT, float] = {}
        self._tag = tag

    def set_tag(self, tag: str | None) -> None:
        self._tag = tag

    def register(
        self,
        hook_type: HookEnumT,
        hook_func: AsyncCallable[..., None],
        once: bool = True,
    ) -> None:
        runner = HookRunner(hook_type, hook_func, once)
        self._hooks[hook_type].append(runner)

    def get_evoke_time(self, hook_type: HookEnumT) -> float:
        return self._stamps.get(hook_type, -1)

    async def emit(
        self,
        hook_type: HookEnumT,
        wait: bool = False,
        /,
        *,
        args: tuple | None = None,
        kwargs: dict[str, Any] | None = None,
    ) -> None:
        self._stamps[hook_type] = time.time_ns() / 1e9
        args = args if args is not None else ()
        kwargs = kwargs if kwargs is not None else {}
        logger = LoggerCtx().get()

        msg = f"<{hook_type}>（{wait = }）"  # noqa: E251, E202
        if self._tag:
            msg = f"开始 {self._tag} 的 hook: {msg}"
        else:
            msg = f"开始 hook: {msg}"
        logger.debug(msg)

        tasks = [
            asyncio.create_task(runner.run(*args, **kwargs))
            for runner in self._hooks[hook_type]
        ]
        if wait and len(tasks):
            await asyncio.wait(tasks)


class Hookable(Generic[HookEnumT]):
    def __init__(self, hook_type: type[HookEnumT], tag: str | None = None):
        super().__init__()
        self._hook_bus = HookBus[HookEnumT](hook_type, tag)

    def on(
        self, *periods: HookEnumT
    ) -> Callable[[AsyncCallable[P, None]], AsyncCallable[P, None]]:
        """注册一个 hook

        :param periods: 要绑定的 hook 类型
        :return: 装饰器
        """

        def wrapped(func: AsyncCallable[P, None]) -> AsyncCallable[P, None]:
            for type in periods:
                self._hook_bus.register(type, func)
            return func

        return wrapped



from enum import Enum
from typing import Iterable, final

from .._hook import HookBus
from ..handle.base import EventHandler
from ..handle.process import Flow
from ..typ import AsyncCallable, Callable, P, SingletonBetterABCMeta, abstractattr
from .ipc import AsyncShare, SyncShare


class PluginLifeSpan(Enum):
    """插件生命周期的枚举"""

    INITED = "i"


class Plugin(metaclass=SingletonBetterABCMeta):
    """插件基类

    子类需要把以下标记为抽象属性的属性，按 :func:`.abstractattr` 的要求实现，其余属性可选实现。
    """

    version: str = abstractattr()
    """插件版本（抽象属性）

       :meta hide-value:
    """

    shares: Iterable[SyncShare | AsyncShare] = ()
    """插件共享对象

       :meta hide-value:
    """

    funcs: Iterable[Callable] = ()
    """插件导出函数

       :meta hide-value:
    """

    flows: Iterable[Flow] = ()
    """插件处理流

       :meta hide-value:
    """

    desc: str = ""
    """插件的短描述信息

       :meta hide-value:
    """

    docs: str = ""
    """插件的长描述信息

       :meta hide-value:
    """

    keywords: list[str] = []
    """插件的分类关键词

       :meta hide-value:
    """

    url: str = ""
    """插件的项目主页

       :meta hide-value:
    """

    author: str = ""
    """插件作者

       :meta hide-value:
    """

    _hook_bus: HookBus[PluginLifeSpan]

    def __init__(self) -> None:
        self.name: str
        self.handlers: tuple[EventHandler, ...]
        self._built: bool = False

    @classmethod
    def __init_hook_bus__(cls, p_name: str | None = None) -> None:
        if not hasattr(cls, "_hook_bus"):
            cls._hook_bus = HookBus[PluginLifeSpan](PluginLifeSpan)
        cls._hook_bus.set_tag(p_name)

    @final
    def __plugin_build__(self, name: str) -> None:
        if not self._built:
            self.name = name
            self.__init_hook_bus__(name)
            self.handlers = tuple(EventHandler(self, flow) for flow in self.flows)
            self._built = True

    @classmethod
    def on(
        cls, *periods: PluginLifeSpan
    ) -> Callable[[AsyncCallable[P, None]], AsyncCallable[P, None]]:
        """注册一个 hook

        :param periods: 要绑定的 hook 类型
        :return: 装饰器
        """
        cls.__init_hook_bus__()

        def wrapped(func: AsyncCallable[P, None]) -> AsyncCallable[P, None]:
            for type in periods:
                cls._hook_bus.register(type, func)
            return func

        return wrapped



from os import PathLike
from typing import Any, Iterable, Literal, Optional, cast

from melobot.adapter import (
    AbstractEchoFactory,
    AbstractEventFactory,
    AbstractOutputFactory,
)
from melobot.adapter import Adapter as RootAdapter
from melobot.adapter import content as mc
from melobot.adapter.content import Content
from melobot.adapter.model import ActionHandle, EchoT
from melobot.ctx import Context
from melobot.exceptions import AdapterError
from melobot.handle import try_get_event
from melobot.typ import AsyncCallable

from ..const import PROTOCOL_IDENTIFIER, P
from ..io.base import BaseIO
from ..io.packet import EchoPacket, InPacket, OutPacket
from . import action as ac
from . import echo as ec
from . import event as ev
from . import segment as se


class EventFactory(AbstractEventFactory[InPacket, ev.Event]):
    async def create(self, packet: InPacket) -> ev.Event:
        return ev.Event.resolve(packet.data)


class OutputFactory(AbstractOutputFactory[OutPacket, ac.Action]):
    async def create(self, action: ac.Action) -> OutPacket:
        return OutPacket(
            data=action.flatten(),
            action_type=action.type,
            action_params=action.params,
            echo_id=action.id if action.need_echo else None,
        )


class EchoFactory(AbstractEchoFactory[EchoPacket, ec.Echo]):
    async def create(self, packet: EchoPacket) -> ec.Echo | None:
        if packet.noecho:
            return None
        return ec.Echo.resolve(action_type=packet.action_type, **packet.data)


class EchoRequireCtx(Context[bool]):
    def __init__(self) -> None:
        super().__init__("ONEBOT_V11_ECHO_STATUS", LookupError)


class Adapter(
    RootAdapter[EventFactory, OutputFactory, EchoFactory, ac.Action, BaseIO, BaseIO]
):
    def __init__(self) -> None:
        super().__init__(
            PROTOCOL_IDENTIFIER, EventFactory(), OutputFactory(), EchoFactory()
        )

    async def call_output(self, action: ac.Action) -> tuple[ActionHandle, ...]:
        """输出行为的底层方法

        :param action: 行为对象
        :return: :class:`.ActionHandle` 元组
        """
        if EchoRequireCtx().try_get():
            action.need_echo = True
        return await super().call_output(action)

    def with_echo(
        self, func: AsyncCallable[P, tuple[ActionHandle[EchoT | None], ...]]
    ) -> AsyncCallable[P, tuple[ActionHandle[EchoT], ...]]:
        async def wrapped_api(
            *args: P.args, **kwargs: P.kwargs
        ) -> tuple[ActionHandle[EchoT], ...]:
            with EchoRequireCtx().unfold(True):
                handles = await func(*args, **kwargs)
            return cast(tuple[ActionHandle[EchoT], ...], handles)

        return wrapped_api

    async def __send_text__(
        self, text: str
    ) -> tuple[ActionHandle[ec.SendMsgEcho | None], ...]:
        return await self.send(text)

    async def __send_media__(
        self,
        name: str,
        raw: bytes | None = None,
        url: str | None = None,
        mimetype: str | None = None,
    ) -> tuple[ActionHandle[ec.SendMsgEcho | None], ...]:
        """通用媒体发送方法"""
        return await self.send(
            se.contents_to_segs(
                [mc.MediaContent(name=name, url=url, raw=raw, mimetype=mimetype)]
            )[0]
        )

    async def __send_image__(
        self,
        name: str,
        raw: bytes | None = None,
        url: str | None = None,
        mimetype: str | None = None,
    ) -> tuple[ActionHandle[ec.SendMsgEcho | None], ...]:
        return await self.send(
            se.contents_to_segs(
                [mc.ImageContent(name=name, url=url, raw=raw, mimetype=mimetype)]
            )[0]
        )

    async def __send_audio__(
        self,
        name: str,
        raw: bytes | None = None,
        url: str | None = None,
        mimetype: str | None = None,
    ) -> tuple[ActionHandle[ec.SendMsgEcho | None], ...]:
        return await self.send(
            se.contents_to_segs(
                [mc.AudioContent(name=name, url=url, raw=raw, mimetype=mimetype)]
            )[0]
        )

    async def __send_voice__(
        self,
        name: str,
        raw: bytes | None = None,
        url: str | None = None,
        mimetype: str | None = None,
    ) -> tuple[ActionHandle[ec.SendMsgEcho | None], ...]:
        return await self.send(
            se.contents_to_segs(
                [mc.VoiceContent(name=name, url=url, raw=raw, mimetype=mimetype)]
            )[0]
        )

    async def __send_video__(
        self,
        name: str,
        raw: bytes | None = None,
        url: str | None = None,
        mimetype: str | None = None,
    ) -> tuple[ActionHandle[ec.SendMsgEcho | None], ...]:
        return await self.send(
            se.contents_to_segs(
                [mc.VideoContent(name=name, url=url, raw=raw, mimetype=mimetype)]
            )[0]
        )

    async def __send_file__(
        self, name: str, path: str | PathLike[str]
    ) -> tuple[ActionHandle[ec.SendMsgEcho | None], ...]:
        return await self.send(
            se.contents_to_segs([mc.FileContent(name=name, flag=str(path))])[0]
        )

    async def __send_refer__(
        self, event: ev.RootEvent, contents: ev.Sequence[Content] | None = None
    ) -> tuple[ActionHandle[ec.SendMsgEcho | None], ...]:
        if not isinstance(event, ev.MessageEvent):
            raise AdapterError(
                f"提供的事件不是 {ev.MessageEvent.__qualname__} 类型，无法用于发送 refer 消息"
            )

        segs = se.contents_to_segs(list(contents)) if contents else []
        segs.insert(0, se.ReplySegment(str(event.message_id)))
        if isinstance(event, ev.GroupMessageEvent):
            return await self.send_custom(segs, group_id=event.group_id)
        return await self.send_custom(segs, user_id=event.user_id)

    async def __send_resource__(
        self, name: str, url: str
    ) -> tuple[ActionHandle[ec.SendMsgEcho | None], ...]:
        return await self.send(se.contents_to_segs([mc.ResourceContent(name, url)])[0])

    async def send(
        self, msgs: str | se.Segment | Iterable[se.Segment] | dict | Iterable[dict]
    ) -> tuple[ActionHandle[ec.SendMsgEcho | None], ...]:
        event = try_get_event()
        if not isinstance(event, ev.MessageEvent):
            raise AdapterError(
                f"当前上下文中不存在事件，或事件不为 {ev.MessageEvent.__qualname__} 类型，无法发送消息"
            )

        if isinstance(event, ev.GroupMessageEvent):
            return await self.send_custom(msgs, group_id=event.group_id)
        return await self.send_custom(msgs, user_id=event.user_id)

    async def send_reply(
        self, msgs: str | se.Segment | Iterable[se.Segment] | dict | Iterable[dict]
    ) -> tuple[ActionHandle[ec.SendMsgEcho | None], ...]:
        event = try_get_event()
        if not isinstance(event, ev.MessageEvent):
            raise AdapterError(
                f"当前上下文中不存在事件，或事件不为 {ev.MessageEvent.__qualname__} 类型，无法发送消息"
            )

        kwargs: dict[str, Any] = {
            "msgs": ac.msgs_to_dicts(se.ReplySegment(str(event.message_id)))
            + ac.msgs_to_dicts(msgs)
        }
        if isinstance(event, ev.GroupMessageEvent):
            kwargs["group_id"] = event.group_id
        else:
            kwargs["user_id"] = event.user_id
        return await self.call_output(ac.SendMsgAction(**kwargs))

    async def send_custom(
        self,
        msgs: str | se.Segment | Iterable[se.Segment] | dict | Iterable[dict],
        user_id: Optional[int] = None,
        group_id: Optional[int] = None,
    ) -> tuple[ActionHandle[ec.SendMsgEcho | None], ...]:
        return await self.call_output(ac.SendMsgAction(msgs, user_id, group_id))

    async def send_forward(
        self, msgs: Iterable[se.NodeSegment]
    ) -> tuple[ActionHandle[ec.SendForwardMsgEcho | None], ...]:
        event = try_get_event()
        if not isinstance(event, ev.MessageEvent):
            raise AdapterError(
                f"当前上下文中不存在事件，或事件不为 {ev.MessageEvent.__qualname__} 类型，无法发送消息"
            )

        if isinstance(event, ev.GroupMessageEvent):
            return await self.send_forward_custom(msgs, group_id=event.group_id)
        return await self.send_forward_custom(msgs, user_id=event.user_id)

    async def send_forward_custom(
        self,
        msgs: Iterable[se.NodeSegment] | Iterable[dict],
        user_id: Optional[int] = None,
        group_id: Optional[int] = None,
    ) -> tuple[ActionHandle[ec.SendForwardMsgEcho | None], ...]:
        return await self.call_output(ac.SendForwardMsgAction(msgs, user_id, group_id))

    async def delete_msg(
        self, msg_id: int
    ) -> tuple[ActionHandle[ec.EmptyEcho | None], ...]:
        return await self.call_output(ac.DeleteMsgAction(msg_id))

    async def get_msg(
        self, msg_id: int
    ) -> tuple[ActionHandle[ec.GetMsgEcho | None], ...]:
        return await self.call_output(ac.GetMsgAction(msg_id))

    async def get_forward_msg(
        self, forward_id: str
    ) -> tuple[ActionHandle[ec.GetForwardMsgEcho | None], ...]:
        return await self.call_output(ac.GetForwardMsgAction(forward_id))

    async def send_like(
        self, user_id: int, times: int = 1
    ) -> tuple[ActionHandle[ec.EmptyEcho | None], ...]:
        return await self.call_output(ac.SendLikeAction(user_id, times))

    async def set_group_kick(
        self, group_id: int, user_id: int, later_reject: bool = False
    ) -> tuple[ActionHandle[ec.EmptyEcho | None], ...]:
        return await self.call_output(
            ac.SetGroupKickAction(group_id, user_id, later_reject)
        )

    async def set_group_ban(
        self, group_id: int, user_id: int, duration: int = 30 * 60
    ) -> tuple[ActionHandle[ec.EmptyEcho | None], ...]:
        return await self.call_output(ac.SetGroupBanAction(group_id, user_id, duration))

    async def set_group_anonymous_ban(
        self,
        group_id: int,
        anonymous: ac.SetGroupAnonymousBanAction.AnonymousDict,
        anonymous_flag: str,
        duration: int = 30 * 60,
    ) -> tuple[ActionHandle[ec.EmptyEcho | None], ...]:
        return await self.call_output(
            ac.SetGroupAnonymousBanAction(group_id, anonymous, anonymous_flag, duration)
        )

    async def set_group_whole_ban(
        self, group_id: int, enable: bool = True
    ) -> tuple[ActionHandle[ec.EmptyEcho | None], ...]:
        return await self.call_output(ac.SetGroupWholeBanAction(group_id, enable))

    async def set_group_admin(
        self, group_id: int, enable: bool = True
    ) -> tuple[ActionHandle[ec.EmptyEcho | None], ...]:
        return await self.call_output(ac.SetGroupAdminAction(group_id, enable))

    async def set_group_anonymous(
        self, group_id: int, enable: bool = True
    ) -> tuple[ActionHandle[ec.EmptyEcho | None], ...]:
        return await self.call_output(ac.SetGroupAnonymousAction(group_id, enable))

    async def set_group_card(
        self, group_id: int, user_id: int, card: str = ""
    ) -> tuple[ActionHandle[ec.EmptyEcho | None], ...]:
        return await self.call_output(ac.SetGroupCardAction(group_id, user_id, card))

    async def set_group_name(
        self, group_id: int, name: str
    ) -> tuple[ActionHandle[ec.EmptyEcho | None], ...]:
        return await self.call_output(ac.SetGroupNameAction(group_id, name))

    async def set_group_leave(
        self, group_id: int, is_dismiss: bool = False
    ) -> tuple[ActionHandle[ec.EmptyEcho | None], ...]:
        return await self.call_output(ac.SetGroupLeaveAction(group_id, is_dismiss))

    async def set_group_special_title(
        self, group_id: int, user_id: int, title: str = "", duration: int = -1
    ) -> tuple[ActionHandle[ec.EmptyEcho | None], ...]:
        return await self.call_output(
            ac.SetGroupSpecialTitleAction(group_id, user_id, title, duration)
        )

    async def set_friend_add_request(
        self, add_flag: str, approve: bool = True, remark: str = ""
    ) -> tuple[ActionHandle[ec.EmptyEcho | None], ...]:
        return await self.call_output(
            ac.SetFriendAddRequestAction(add_flag, approve, remark)
        )

    async def set_group_add_request(
        self,
        add_flag: str,
        add_type: Literal["add", "invite"],
        approve: bool = True,
        reason: str = "",
    ) -> tuple[ActionHandle[ec.EmptyEcho | None], ...]:
        return await self.call_output(
            ac.SetGroupAddRequestAction(add_flag, add_type, approve, reason)
        )

    async def get_login_info(self) -> tuple[ActionHandle[ec.GetLoginInfoEcho], ...]:
        return await self.call_output(ac.GetLoginInfoAction())

    async def get_stranger_info(
        self, user_id: int, no_cache: bool = False
    ) -> tuple[ActionHandle[ec.GetStrangerInfoEcho], ...]:
        return await self.call_output(ac.GetStrangerInfoAction(user_id, no_cache))

    async def get_friend_list(self) -> tuple[ActionHandle[ec.GetFriendListEcho], ...]:
        return await self.call_output(ac.GetFriendlistAction())

    async def get_group_info(
        self, group_id: int, no_cache: bool = False
    ) -> tuple[ActionHandle[ec.GetGroupInfoEcho], ...]:
        return await self.call_output(ac.GetGroupInfoAction(group_id, no_cache))

    async def get_group_list(self) -> tuple[ActionHandle[ec.GetGroupListEcho], ...]:
        return await self.call_output(ac.GetGrouplistAction())

    async def get_group_member_info(
        self, group_id: int, user_id: int, no_cache: bool = False
    ) -> tuple[ActionHandle[ec.GetGroupMemberInfoEcho], ...]:
        return await self.call_output(
            ac.GetGroupMemberInfoAction(group_id, user_id, no_cache)
        )

    async def get_group_member_list(
        self, group_id: int
    ) -> tuple[ActionHandle[ec.GetGroupMemberListEcho], ...]:
        return await self.call_output(ac.GetGroupMemberlistAction(group_id))

    async def get_group_honor_info(
        self,
        group_id: int,
        type: Literal[
            "talkative", "performer", "legend", "strong_newbie", "emotion", "all"
        ],
    ) -> tuple[ActionHandle[ec.GetGroupHonorInfoEcho], ...]:
        return await self.call_output(ac.GetGroupHonorInfoAction(group_id, type))

    async def get_cookies(
        self, domain: str = ""
    ) -> tuple[ActionHandle[ec.GetCookiesEcho], ...]:
        return await self.call_output(ac.GetCookiesAction(domain))

    async def get_csrf_token(self) -> tuple[ActionHandle[ec.GetCsrfTokenEcho], ...]:
        return await self.call_output(ac.GetCsrfTokenAction())

    async def get_credentials(
        self, domain: str = ""
    ) -> tuple[ActionHandle[ec.GetCredentialsEcho], ...]:
        return await self.call_output(ac.GetCredentialsAction(domain))

    async def get_record(
        self, file: str, out_format: str
    ) -> tuple[ActionHandle[ec.GetRecordEcho], ...]:
        return await self.call_output(ac.GetRecordAction(file, out_format))

    async def get_image(self, file: str) -> tuple[ActionHandle[ec.GetImageEcho], ...]:
        return await self.call_output(ac.GetImageAction(file))

    async def can_send_image(self) -> tuple[ActionHandle[ec.CanSendImageEcho], ...]:
        return await self.call_output(ac.CanSendImageAction())

    async def can_send_record(self) -> tuple[ActionHandle[ec.CanSendRecordEcho], ...]:
        return await self.call_output(ac.CanSendRecordAction())

    async def get_status(self) -> tuple[ActionHandle[ec.GetStatusEcho], ...]:
        return await self.call_output(ac.GetStatusAction())

    async def get_version_info(self) -> tuple[ActionHandle[ec.GetVersionInfoEcho], ...]:
        return await self.call_output(ac.GetVersionInfoAction())

    async def set_restart(
        self, delay: int = 0
    ) -> tuple[ActionHandle[ec.EmptyEcho | None], ...]:
        return await self.call_output(ac.SetRestartAction(delay))

    async def clean_cache(self) -> tuple[ActionHandle[ec.EmptyEcho | None], ...]:
        return await self.call_output(ac.CleanCacheAction())



from os import PathLike
from typing import Sequence

from ..ctx import EventBuildInfoCtx
from .content import Content
from .model import ActionHandle, Event

_CTX = EventBuildInfoCtx()


async def send_text(text: str) -> tuple[ActionHandle, ...]:
    """通用文本输出方法"""
    return await _CTX.get().adapter.__send_text__(text)


async def send_media(
    name: str,
    raw: bytes | None = None,
    url: str | None = None,
    mimetype: str | None = None,
) -> tuple[ActionHandle, ...]:
    """通用媒体内容输出方法"""
    return await _CTX.get().adapter.__send_media__(name, raw, url, mimetype)


async def send_image(
    name: str,
    raw: bytes | None = None,
    url: str | None = None,
    mimetype: str | None = None,
) -> tuple[ActionHandle, ...]:
    """通用图像内容输出方法"""
    return await _CTX.get().adapter.__send_image__(name, raw, url, mimetype)


async def send_audio(
    name: str,
    raw: bytes | None = None,
    url: str | None = None,
    mimetype: str | None = None,
) -> tuple[ActionHandle, ...]:
    """通用音频内容输出方法"""
    return await _CTX.get().adapter.__send_audio__(name, raw, url, mimetype)


async def send_voice(
    name: str,
    raw: bytes | None = None,
    url: str | None = None,
    mimetype: str | None = None,
) -> tuple[ActionHandle, ...]:
    """通用语音内容输出方法"""
    return await _CTX.get().adapter.__send_voice__(name, raw, url, mimetype)


async def send_video(
    name: str,
    raw: bytes | None = None,
    url: str | None = None,
    mimetype: str | None = None,
) -> tuple[ActionHandle, ...]:
    """通用视频内容输出方法"""
    return await _CTX.get().adapter.__send_video__(name, raw, url, mimetype)


async def send_file(name: str, path: str | PathLike[str]) -> tuple[ActionHandle, ...]:
    """通用文件输出方法"""
    return await _CTX.get().adapter.__send_file__(name, path)


async def send_refer(
    event: Event, contents: Sequence[Content] | None = None
) -> tuple[ActionHandle, ...]:
    """通用过往事件引用输出方法"""
    return await _CTX.get().adapter.__send_refer__(event, contents)


async def send_resource(name: str, url: str) -> tuple[ActionHandle, ...]:
    """通用其他资源输出方法"""
    return await _CTX.get().adapter.__send_resource__(name, url)



from .. import __version__
from .adapter import Adapter, EchoRequireCtx
from .adapter.action import Action
from .adapter.echo import Echo
from .adapter.event import Event
from .adapter.segment import Segment
from .const import PROTOCOL_IDENTIFIER
from .handle import (
    msg_session,
    on_at_qq,
    on_command,
    on_contain_match,
    on_end_match,
    on_event,
    on_full_match,
    on_message,
    on_meta,
    on_notice,
    on_regex_match,
    on_request,
    on_start_match,
)
from .io import ForwardWebSocketIO, HttpIO, ReverseWebSocketIO
from .utils import GroupRole, LevelRole, ParseArgs



import inspect
import warnings
from abc import ABCMeta, abstractmethod
from enum import Enum
from functools import wraps
from typing import Any, Awaitable, Callable, Protocol, Sequence, cast

from beartype import BeartypeConf as _BeartypeConf
from beartype.door import is_bearable as _is_type
from beartype.door import is_subhint
from typing_extensions import ParamSpec, Self, TypeGuard, TypeVar

__all__ = (
    "T",
    "P",
    "AsyncCallable",
    "is_type",
    "is_subhint",
    "HandleLevel",
    "LogicMode",
    "BetterABCMeta",
    "BetterABC",
    "abstractattr",
    "abstractmethod",
    "VoidType",
)

#: 泛型 T，无约束
T = TypeVar("T", default=Any)
#: 泛型 T_co，协变无约束
T_co = TypeVar("T_co", covariant=True, default=Any)
#: :obj:`~typing.ParamSpec` 泛型 P，无约束
P = ParamSpec("P", default=Any)


class AsyncCallable(Protocol[P, T_co]):
    """用法：AsyncCallable[P, T]

    是该类型的等价形式：Callable[P, Awaitable[T]]
    """

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> Awaitable[T_co]: ...


_DEFAULT_BEARTYPE_CONF = _BeartypeConf(is_pep484_tower=True)


def is_type(obj: T, hint: type[Any]) -> TypeGuard[T]:
    """检查 `obj` 是否是类型注解 `hint` 所表示的类型

    :param obj: 任意对象
    :param hint: 任意类型注解
    :return: 布尔值
    """
    ret = _is_type(obj, hint, conf=_DEFAULT_BEARTYPE_CONF)
    return ret


class HandleLevel(float, Enum):
    """事件处理流优先级枚举类型"""

    MAX = 1 << 6
    ULTRA_HIGH = 1 << 5
    HIGH = 1 << 4
    NORMAL = 1 << 3
    LOW = 1 << 2
    ULTRA_LOW = 1 << 1
    MIN = 1


class LogicMode(Enum):
    """逻辑模式枚举类型"""

    AND = 1
    OR = 2
    NOT = 3
    XOR = 4

    @classmethod
    def calc(cls, logic: "LogicMode", v1: Any, v2: Any = None) -> bool:
        """将两个值使用指定逻辑模式运算

        :param logic: 逻辑模式
        :param v1: 值 1
        :param v2: 值 2
        :return: 布尔值
        """
        if logic == LogicMode.AND:
            return (v1 and v2) if v2 is not None else bool(v1)  # type: ignore[no-any-return]
        if logic == LogicMode.OR:
            return (v1 or v2) if v2 is not None else bool(v1)  # type: ignore[no-any-return]
        if logic == LogicMode.NOT:
            return not v1
        return (v1 ^ v2) if v2 is not None else bool(v1)  # type: ignore[no-any-return]

    @classmethod
    def short_calc(
        cls, logic: "LogicMode", v1: Callable[[], Any], v2: Callable[[], Any]
    ) -> bool:
        """与 :func:`calc` 功能类似，但运算支持短路

        :param logic: 逻辑模式
        :param v1: 生成值 1 的可调用对象
        :param v2: 生成值 2 的可调用对象
        :return: 布尔值
        """
        if logic == LogicMode.AND:
            return (v1() and v2()) if v2 is not None else bool(v1())  # type: ignore[no-any-return]
        if logic == LogicMode.OR:
            return (v1() or v2()) if v2 is not None else bool(v1())  # type: ignore[no-any-return]
        if logic == LogicMode.NOT:
            return not v1()
        return (v1() ^ v2()) if v2 is not None else bool(v1())  # type: ignore[no-any-return]

    @classmethod
    async def async_short_calc(
        cls, logic: "LogicMode", v1: AsyncCallable[[], Any], v2: AsyncCallable[[], Any]
    ) -> bool:
        """与 :func:`short_calc` 功能类似，但运算支持异步

        :param logic: 逻辑模式
        :param v1: 生成值 1 的异步可调用对象
        :param v2: 生成值 2 的异步可调用对象
        :return: 布尔值
        """
        if logic == LogicMode.AND:
            res = (await v1() and await v2()) if v2 is not None else bool(await v1())
            return res  # type: ignore[no-any-return]
        if logic == LogicMode.OR:
            res = (await v1() or await v2()) if v2 is not None else bool(await v1())
            return res  # type: ignore[no-any-return]
        if logic == LogicMode.NOT:
            return not await v1()
        res = (await v1() ^ await v2()) if v2 is not None else bool(await v1())
        return res  # type: ignore[no-any-return]

    @classmethod
    def seq_calc(cls, logic: "LogicMode", values: list[Any]) -> bool:
        """使用指定的逻辑模式，对值序列进行运算

        .. code:: python

            # 操作等价与：True and False and True
            LogicMode.seq_calc(LogicMode.AND, [True, False, True])

        :param logic: 逻辑模式
        :param values: 值序列
        :return: 布尔值
        """
        if len(values) <= 0:
            return False
        if len(values) <= 1:
            return bool(values[0])

        idx = 0
        res: bool
        while idx < len(values):
            if idx == 0:
                res = cls.calc(logic, values[idx], values[idx + 1])
                idx += 1
            else:
                res = cls.calc(logic, res, values[idx])
            idx += 1
        return res

    @classmethod
    def short_seq_calc(
        cls, logic: "LogicMode", getters: Sequence[Callable[[], Any]]
    ) -> bool:
        """与 :func:`seq_calc` 功能类似，但运算支持短路

        :param logic: 逻辑模式
        :param getters: 一组获取值的可调用对象
        :return: 布尔值
        """
        if len(getters) <= 0:
            return False
        if len(getters) <= 1:
            return bool(getters[0]())

        idx = 0
        res: bool
        while idx < len(getters):
            if idx == 0:
                res = cls.short_calc(logic, getters[idx], getters[idx + 1])
                idx += 1
            else:
                res = cls.short_calc(logic, lambda: res, getters[idx])
            idx += 1
        return res

    @classmethod
    async def async_short_seq_calc(
        cls, logic: "LogicMode", getters: Sequence[AsyncCallable[[], Any]]
    ) -> bool:
        """与 :func:`short_seq_calc` 功能类似，但运算支持异步

        :param logic: 逻辑模式
        :param getters: 一组获取值的异步可调用对象
        :return: 布尔值
        """
        if len(getters) <= 0:
            return False
        if len(getters) <= 1:
            return bool(await getters[0]())

        idx = 0
        res: bool
        while idx < len(getters):
            if idx == 0:
                res = await cls.async_short_calc(logic, getters[idx], getters[idx + 1])
                idx += 1
            else:

                async def res_getter() -> bool:
                    return res

                res = await cls.async_short_calc(logic, res_getter, getters[idx])
            idx += 1
        return res


def abstractattr(obj: Callable[[Any], T] | None = None) -> T:
    """抽象属性

    与 `abstractproperty` 相比更灵活，`abstractattr` 不关心你以何种形式定义属性。只要子类在实例化后，该属性存在，即认为合法。

    但注意它必须与 :class:`BetterABC` 或 :class:`BetterABCMeta` 或 :class:`.SingletonBetterABCMeta` 配合使用

    这意味着可以在类层级、实例层级定义属性，或使用 `property` 定义属性：

    .. code:: python

        class A(BetterABC):
            foo: int = abstractattr()  # 声明为抽象属性

            # 或者使用装饰器的形式声明，这与上面是等价的
            @abstractattr
            def bar(self) -> int: ...

        # 以下实现都是合法的：

        class B(A):
            foo = 2
            bar = 4

        class C(A):
            foo = 3
            def __init__(self) -> None:
                self.bar = 5

        class D(A):
            def __init__(self) -> None:
                self.foo = 8

            @property
            def bar(self) -> int:
                return self.foo + 10
    """
    _obj = cast(Any, obj)
    if obj is None:
        _obj = BetterABCMeta.DummyAttribute()
    setattr(_obj, "__is_abstract_attribute__", True)
    return cast(T, _obj)


class BetterABCMeta(ABCMeta):
    """更好的抽象元类，兼容 `ABCMeta` 的所有功能，但是额外支持 :func:`abstractattr`"""

    class DummyAttribute: ...

    def __call__(cls: type[T], *args: Any, **kwargs: Any) -> T:
        instance = ABCMeta.__call__(cls, *args, **kwargs)
        lack_attrs = set()
        for name in dir(instance):
            try:
                attr = getattr(instance, name)
            except Exception:
                if not isinstance(getattr(instance.__class__, name), property):
                    raise

            if getattr(attr, "__is_abstract_attribute__", False):
                lack_attrs.add(name)
            if inspect.iscoroutine(attr):
                attr.close()

        if lack_attrs:
            raise NotImplementedError(
                "Can't instantiate abstract class {} with"
                " abstract attributes: {}".format(cls.__name__, ", ".join(lack_attrs))
            )
        return cast(T, instance)


class BetterABC(metaclass=BetterABCMeta):
    """更好的抽象类，兼容 `ABC` 的所有功能，但是额外支持 :func:`abstractattr`"""

    __slots__ = ()


class SingletonMeta(type):
    """单例元类

    相比单例装饰器，可以自动保证所有子类都为单例
    """

    __instances__: dict[type, Any] = {}

    def __call__(cls: type[T], *args: Any, **kwargs: Any) -> T:
        if cls not in SingletonMeta.__instances__:
            SingletonMeta.__instances__[cls] = cast(
                T, super(SingletonMeta, cls).__call__(*args, **kwargs)  # type: ignore[misc]
            )
        return cast(T, SingletonMeta.__instances__[cls])


class SingletonBetterABCMeta(BetterABCMeta):
    """单例抽象元类

    相比普通的抽象元类，还可以自动保证所有子类都为单例
    """

    __instances__: dict[type, Any] = {}

    def __call__(cls: type[T], *args: Any, **kwargs: Any) -> T:
        mcls = SingletonBetterABCMeta
        if cls not in mcls.__instances__:
            mcls.__instances__[cls] = BetterABCMeta.__call__(cls, *args, **kwargs)
        return cast(T, mcls.__instances__[cls])


class Markable:
    """可标记对象

    无需直接实例化，而是用作接口在其他类中继承
    """

    def __init__(self) -> None:
        self._flags: dict[str, dict[str, Any]] = {}

    def flag_mark(self, namespace: str, flag_name: str, val: Any = None) -> None:
        """在 `namespace` 命名空间中设置 `flag_name` 标记，值为 `val`

        注：不同的对象并不共享 `namespace`，`namespace` 只适用于单个对象

        :param namespace: 命名空间
        :param flag_name: 标记名
        :param val: 标记值
        """
        self._flags.setdefault(namespace, {})

        if flag_name in self._flags[namespace].keys():
            raise ValueError(
                f"标记失败。对象的命名空间 {namespace} 中已存在名为 {flag_name} 的标记"
            )

        self._flags[namespace][flag_name] = val

    def flag_check(self, namespace: str, flag_name: str, val: Any = None) -> bool:
        """检查 `namespace` 命名空间中 `flag_name` 标记值是否为 `val`

        注：不同的对象并不共享 `namespace`，`namespace` 只适用于单个对象

        :param namespace: 命名空间
        :param flag_name: 标记名
        :param val: 标记值
        :return: 是否通过检查
        """
        if self._flags.get(namespace) is None:
            return False
        if flag_name not in self._flags[namespace].keys():
            return False
        flag = self._flags[namespace][flag_name]

        if val is None:
            return flag is None
        return cast(bool, flag == val)


class AttrsReprable:
    def __repr__(self) -> str:
        attrs = ", ".join(
            f"{k}={repr(v)}" for k, v in self.__dict__.items() if not k.startswith("_")
        )
        return f"{self.__class__.__name__}({attrs})"


class Locatable:
    def __new__(cls, *_args: Any, **_kwargs: Any) -> Self:
        obj = super().__new__(cls)
        obj.__obj_location__ = obj._init_location()  # type: ignore[attr-defined]
        return obj

    def __init__(self) -> None:
        self.__obj_location__: tuple[str, str, int]

    @staticmethod
    def _init_location() -> tuple[str, str, int]:
        frame = inspect.currentframe()
        while frame:
            if frame.f_code.co_name == "<module>":
                return (
                    frame.f_globals["__name__"],
                    frame.f_globals["__file__"],
                    frame.f_lineno,
                )
            frame = frame.f_back

        return (
            "<unknown module>",
            "<unknown file>",
            -1,
        )

    @property
    def __obj_module__(self) -> str:
        return self.__obj_location__[0]

    @property
    def __obj_file__(self) -> str:
        return self.__obj_location__[1]

    @property
    def __obj_line__(self) -> int:
        return self.__obj_location__[2]


class VoidType(Enum):
    """空类型，需要区别于 `None` 时使用

    .. code:: python

        # 有些时候 `None` 也是合法值，因此需要一个额外的哨兵值：
        def foo(val: Any | VoidType = VoidType.VOID) -> None:
            ...
    """

    VOID = type("_VOID", (), {})


def deprecated(msg: str) -> Callable[[Callable[P, T]], Callable[P, T]]:

    def decorator(func: Callable[P, T]) -> Callable[P, T]:

        @wraps(func)
        def deprecate_wrapped(*args: P.args, **kwargs: P.kwargs) -> T:
            warnings.warn(
                f"使用了弃用函数/方法 {func.__qualname__}: {msg}",
                category=DeprecationWarning,
                stacklevel=2,
            )
            return func(*args, **kwargs)

        return deprecate_wrapped

    return decorator



from __future__ import annotations

from typing import Callable, Generic, final

from ..adapter.model import EventT
from ..typ import BetterABC, abstractmethod


class Rule(BetterABC, Generic[EventT]):
    """会话规则

    对于更复杂的情况，可以继承此类，在子类中结合状态信息实现更动态的会话判断
    """

    @staticmethod
    @final
    def new(meth: Callable[[EventT, EventT], bool]) -> Rule[EventT]:
        """从可调用对象创建一个新的会话规则对象

        :param meth: 用于会话判断的可调用对象
        :return: 会话规则
        """
        return _CustomRule[EventT](meth)

    @abstractmethod
    async def compare(self, e1: EventT, e2: EventT) -> bool:
        """会话判断的方法

        :param e1: 事件 1
        :param e2: 事件 2
        :return: 为 `True` 则在一个会话中，反之亦然
        """
        raise NotImplementedError


class _CustomRule(Rule[EventT]):
    def __init__(self, meth: Callable[[EventT, EventT], bool]) -> None:
        super().__init__()
        self.meth = meth

    async def compare(self, e1: EventT, e2: EventT) -> bool:
        return self.meth(e1, e2)



from typing import ParamSpec, TypeVar

T = TypeVar("T")
V = TypeVar("V")
P = ParamSpec("P")


PROTOCOL_NAME = "OneBot"
PROTOCOL_VERSION = "11"
MODULE_AUTHOR = "Meloland"
PROTOCOL_IDENTIFIER = f"{PROTOCOL_NAME}-v{PROTOCOL_VERSION}@{MODULE_AUTHOR}"



from ..ctx import SessionCtx as _SessionCtx
from .base import Session, SessionStore, enter_session, suspend
from .option import Rule


def get_session_store() -> SessionStore:
    """获取当前上下文中的会话存储

    :return: 会话存储
    """
    return _SessionCtx().get().store


def get_rule() -> Rule | None:
    """获取当前上下文中的会话规则

    :return: 会话规则或空
    """
    return _SessionCtx().get().rule



from contextlib import _AsyncGeneratorContextManager
from functools import wraps
from typing import Callable, cast

from melobot.ctx import Context
from melobot.di import Depends, inject_deps
from melobot.handle import Flow, get_event, no_deps_node
from melobot.session import Rule, Session, enter_session
from melobot.typ import AsyncCallable, HandleLevel, LogicMode, deprecated
from melobot.utils import get_obj_name

from .adapter.event import Event, MessageEvent, MetaEvent, NoticeEvent, RequestEvent
from .utils import check, match
from .utils.abc import Checker, Matcher, ParseArgs, Parser
from .utils.parse import CmdArgFormatter, CmdParser


class ParseArgsCtx(Context[ParseArgs | None]):
    def __init__(self) -> None:
        super().__init__(
            "ONEBOT_V11_PARSE_ARGS", LookupError, "当前上下文中不存在解析参数"
        )


def GetParseArgs() -> ParseArgs:  # pylint: disable=invalid-name
    """获取解析参数

    :return: 解析参数
    """
    return cast(ParseArgs, Depends(ParseArgsCtx().get, recursive=False))


@deprecated(
    "将于 melobot v3.0.0 移除，使用 melobot.protocols.onebot.v11.handle.GetParseArgs 代替"
)
def Args() -> ParseArgs:  # pylint: disable=invalid-name
    """获取解析参数，与 :func:`~.v11.handle.GetParseArgs` 等价

    .. admonition:: 重要提示
        :class: caution

        已弃用。将于 `v3.0.0` 移除，使用 :func:`~.v11.handle.GetParseArgs` 代替

    :return: 解析参数
    """
    return GetParseArgs()


FlowDecorator = Callable[[AsyncCallable[..., bool | None]], Flow]

_DefaultRule = Rule[MessageEvent].new(lambda e1, e2: e1.scope == e2.scope)


def msg_session() -> _AsyncGeneratorContextManager[Session]:
    """消息事件的会话上下文管理器

    展开一个传统意义上的消息会话上下文，这与其他 bot 开发框架中的“会话”语义等价
    """
    return enter_session(_DefaultRule)


def on_event(
    checker: Checker | None | Callable[[Event], bool] = None,
    matcher: Matcher | None = None,
    parser: Parser | None = None,
    priority: HandleLevel = HandleLevel.NORMAL,
    block: bool = False,
    temp: bool = False,
) -> FlowDecorator:

    def flow_getter(func: AsyncCallable[..., bool | None]) -> Flow:
        if not isinstance(checker, Checker) and callable(checker):
            _checker = Checker.new(checker)
        else:
            _checker = cast(Checker, checker)

        func = inject_deps(func)

        @wraps(func)
        async def wrapped() -> bool | None:
            event = cast(Event, get_event())
            status = await _checker.check(event)
            if not status:
                return None

            p_args: ParseArgs | None = None
            if isinstance(event, MessageEvent):
                if matcher:
                    status = await matcher.match(event.text)
                    if not status:
                        return None

                if parser:
                    parse_args = await parser.parse(event.text)
                    if parse_args is None:
                        return None
                    p_args = parse_args

            event.spread = not block
            with ParseArgsCtx().unfold(p_args):
                return await func()

        n = no_deps_node(wrapped)
        n.name = get_obj_name(func, otype="callable")
        return Flow(
            f"OneBotV11Flow[{n.name}]",
            (n,),
            priority=priority,
            temp=temp,
        )

    return flow_getter


def _checker_join(*checkers: Checker | None | Callable[[Event], bool]) -> Checker:
    checker: Checker | None = None
    for c in checkers:
        if c is None:
            continue
        if isinstance(c, Checker):
            checker = checker & c if checker else c
        else:
            checker = checker & Checker.new(c) if checker else Checker.new(c)

    if checker is None:
        raise ValueError("检查器序列不能全为空")
    return checker


def on_message(
    checker: Checker | None | Callable[[MessageEvent], bool] = None,
    matcher: Matcher | None = None,
    parser: Parser | None = None,
    priority: HandleLevel = HandleLevel.NORMAL,
    block: bool = False,
    temp: bool = False,
) -> FlowDecorator:
    return on_event(
        _checker_join(lambda e: e.is_message(), checker),  # type: ignore[arg-type]
        matcher,
        parser,
        priority,
        block,
        temp,
    )


def on_at_qq(
    qid: int | None = None,
    checker: Checker | None | Callable[[MessageEvent], bool] = None,
    matcher: Matcher | None = None,
    parser: Parser | None = None,
    priority: HandleLevel = HandleLevel.NORMAL,
    block: bool = False,
    temp: bool = False,
) -> FlowDecorator:
    return on_message(
        _checker_join(check.AtMsgChecker(qid if qid else "all"), checker),  # type: ignore[arg-type]
        matcher,
        parser,
        priority,
        block,
        temp,
    )


def on_command(
    cmd_start: str | list[str],
    cmd_sep: str | list[str],
    targets: str | list[str],
    fmtters: list[CmdArgFormatter | None] | None = None,
    checker: Checker | None | Callable[[MessageEvent], bool] = None,
    matcher: Matcher | None = None,
    priority: HandleLevel = HandleLevel.NORMAL,
    block: bool = False,
    temp: bool = False,
) -> FlowDecorator:
    return on_message(
        checker,
        matcher,
        CmdParser(cmd_start=cmd_start, cmd_sep=cmd_sep, targets=targets, fmtters=fmtters),
        priority,
        block,
        temp,
    )


def on_start_match(
    target: str | list[str],
    logic_mode: LogicMode = LogicMode.OR,
    checker: Checker | Callable[[MessageEvent], bool] | None = None,
    parser: Parser | None = None,
    priority: HandleLevel = HandleLevel.NORMAL,
    block: bool = False,
    temp: bool = False,
) -> FlowDecorator:
    return on_message(
        checker, match.StartMatcher(target, logic_mode), parser, priority, block, temp
    )


def on_contain_match(
    target: str | list[str],
    logic_mode: LogicMode = LogicMode.OR,
    checker: Checker | Callable[[MessageEvent], bool] | None = None,
    parser: Parser | None = None,
    priority: HandleLevel = HandleLevel.NORMAL,
    block: bool = False,
    temp: bool = False,
) -> FlowDecorator:
    return on_message(
        checker, match.ContainMatcher(target, logic_mode), parser, priority, block, temp
    )


def on_full_match(
    target: str | list[str],
    logic_mode: LogicMode = LogicMode.OR,
    checker: Checker | Callable[[MessageEvent], bool] | None = None,
    parser: Parser | None = None,
    priority: HandleLevel = HandleLevel.NORMAL,
    block: bool = False,
    temp: bool = False,
) -> FlowDecorator:
    return on_message(
        checker, match.FullMatcher(target, logic_mode), parser, priority, block, temp
    )


def on_end_match(
    target: str | list[str],
    logic_mode: LogicMode = LogicMode.OR,
    checker: Checker | Callable[[MessageEvent], bool] | None = None,
    parser: Parser | None = None,
    priority: HandleLevel = HandleLevel.NORMAL,
    block: bool = False,
    temp: bool = False,
) -> FlowDecorator:
    return on_message(
        checker, match.EndMatcher(target, logic_mode), parser, priority, block, temp
    )


def on_regex_match(
    target: str,
    logic_mode: LogicMode = LogicMode.OR,
    checker: Checker | Callable[[MessageEvent], bool] | None = None,
    parser: Parser | None = None,
    priority: HandleLevel = HandleLevel.NORMAL,
    block: bool = False,
    temp: bool = False,
) -> FlowDecorator:
    return on_message(
        checker, match.RegexMatcher(target, logic_mode), parser, priority, block, temp
    )


def on_request(
    checker: Checker | None | Callable[[RequestEvent], bool] = None,
    matcher: Matcher | None = None,
    parser: Parser | None = None,
    priority: HandleLevel = HandleLevel.NORMAL,
    block: bool = False,
    temp: bool = False,
) -> FlowDecorator:
    return on_event(
        _checker_join(lambda e: e.is_request(), checker),  # type: ignore[arg-type]
        matcher,
        parser,
        priority,
        block,
        temp,
    )


def on_notice(
    checker: Checker | None | Callable[[NoticeEvent], bool] = None,
    matcher: Matcher | None = None,
    parser: Parser | None = None,
    priority: HandleLevel = HandleLevel.NORMAL,
    block: bool = False,
    temp: bool = False,
) -> FlowDecorator:
    return on_event(
        _checker_join(lambda e: e.is_notice(), checker),  # type: ignore[arg-type]
        matcher,
        parser,
        priority,
        block,
        temp,
    )


def on_meta(
    checker: Checker | None | Callable[[MetaEvent], bool] = None,
    matcher: Matcher | None = None,
    parser: Parser | None = None,
    priority: HandleLevel = HandleLevel.NORMAL,
    block: bool = False,
    temp: bool = False,
) -> FlowDecorator:
    return on_event(
        _checker_join(lambda e: e.is_meta(), checker),  # type: ignore[arg-type]
        matcher,
        parser,
        priority,
        block,
        temp,
    )



from __future__ import annotations

from asyncio import Lock
from collections import deque
from dataclasses import dataclass
from functools import wraps
from inspect import Parameter, isawaitable, signature, unwrap
from sys import version_info
from types import BuiltinFunctionType, FunctionType, LambdaType
from typing import (
    TYPE_CHECKING,
    Annotated,
    Any,
    Callable,
    Generic,
    Sequence,
    cast,
    get_args,
    get_origin,
)

from .ctx import BotCtx, EventBuildInfoCtx, FlowCtx, LoggerCtx, SessionCtx
from .exceptions import DependBindError, DependInitError
from .typ import (
    AsyncCallable,
    BetterABC,
    P,
    T,
    VoidType,
    abstractmethod,
    is_subhint,
    is_type,
)
from .utils import get_obj_name, to_async

if TYPE_CHECKING:
    from .adapter.base import Adapter


class DependNotMatched(BaseException):
    def __init__(
        self, msg: str, func_name: str, arg_name: str, real_type: type, hint: Any
    ) -> None:
        super().__init__(msg)
        self.func_name = func_name
        self.arg_name = arg_name
        self.real_type = real_type
        self.hint = hint


class Depends:
    def __init__(
        self,
        dep: Callable[[], Any] | AsyncCallable[[], Any] | Depends,
        sub_getter: Callable[[Any], Any] | AsyncCallable[[Any], Any] | None = None,
        cache: bool = False,
        recursive: bool = True,
    ) -> None:
        """初始化一个依赖项

        :param dep: 依赖来源（可调用对象，异步可调用对象，或依赖项）
        :param sub_getter: 子获取器（可调用对象，异步可调用对象或空），在获得依赖之后，于其上继续获取
        :param cache: 是否启用缓存
        :param recursive: 是否启用递归满足（默认启用，如果当前依赖来源中存在依赖项，会被递归满足；关闭可节约性能）
        """
        super().__init__()
        self.ref: Depends | None
        self.getter: AsyncCallable[[], Any] | None

        if isinstance(dep, Depends):
            self.ref = dep
            self.getter = None
        else:
            self.ref = None
            if recursive:
                self.getter = inject_deps(dep)
            else:
                self.getter = to_async(dep)

        if sub_getter is None:
            self.sub_getter = None
        elif recursive:
            self.sub_getter = inject_deps(sub_getter)
        else:
            self.sub_getter = to_async(sub_getter)

        self._lock = Lock() if cache else None
        self._cached: Any = VoidType.VOID

    def __repr__(self) -> str:
        getter_str = f"getter={self.getter}" if self.getter is not None else ""
        ref_str = f"ref={self.ref}" if self.ref is not None else ""
        return f"{self.__class__.__name__}({ref_str if ref_str != '' else getter_str})"

    async def _get(self, dep_scope: dict[Depends, Any]) -> Any:
        if self.getter is not None:
            val = await self.getter()
        else:
            ref = cast(Depends, self.ref)
            val = dep_scope.get(ref, VoidType.VOID)
            if val is VoidType.VOID:
                val = await ref.fulfill(dep_scope)

        if self.sub_getter is not None:
            val = self.sub_getter(val)
            if isawaitable(val):
                val = await val
        return val

    async def fulfill(self, dep_scope: dict[Depends, Any]) -> Any:
        if self._lock is None:
            val = await self._get(dep_scope)
        elif self._cached is not VoidType.VOID:
            val = self._cached
        else:
            async with self._lock:
                if self._cached is VoidType.VOID:
                    self._cached = await self._get(dep_scope)
                val = self._cached

        dep_scope[self] = val
        return val


def _adapter_get(hint: Any) -> "Adapter | None":
    ctx = EventBuildInfoCtx()
    try:
        return ctx.get().adapter
    except ctx.lookup_exc_cls:
        return BotCtx().get().get_adapter(hint)


def _custom_logger_get(hint: Any, data: CustomLogger) -> Any:
    val = LoggerCtx().get()
    if not is_type(val, hint):
        val = data.getter()
    return val


class AutoDepends(Depends):
    def __init__(self, func: Callable, name: str, hint: Any) -> None:
        self.hint = hint
        self.func = func
        self.func_name = get_obj_name(func, otype="callable")
        self.arg_name = name

        if get_origin(hint) is Annotated:
            args = get_args(hint)
            if not len(args):
                raise DependInitError(
                    "可依赖注入的函数若使用 Annotated 注解，必须附加元数据"
                )
            self.metadatas = args
        else:
            self.metadatas = ()

        self.orig_getter: Callable[[], Any] | AsyncCallable[[], Any] | None = None

        if is_subhint(hint, FlowCtx().get_event_type()):
            self.orig_getter = FlowCtx().get_event

        elif is_subhint(hint, BotCtx().get_type()):
            self.orig_getter = BotCtx().get

        elif is_subhint(hint, EventBuildInfoCtx().get_adapter_type()):
            self.orig_getter = cast(Callable[[], Any], lambda h=hint: _adapter_get(h))

        elif is_subhint(hint, LoggerCtx().get_type()):
            self.orig_getter = LoggerCtx().get

        elif is_subhint(hint, FlowCtx().get_store_type()):
            self.orig_getter = FlowCtx().get_store

        elif is_subhint(hint, SessionCtx().get_store_type()):
            self.orig_getter = SessionCtx().get_store

        elif is_subhint(hint, SessionCtx().get_rule_type() | None):
            self.orig_getter = lambda: SessionCtx().get().rule

        for data in self.metadatas:
            if isinstance(data, CustomLogger):
                self.orig_getter = cast(
                    Callable[[], Any], lambda h=hint, d=data: _custom_logger_get(h, d)
                )
                break

        if self.orig_getter is None:
            raise DependInitError(
                f"函数 {self.func_name} 的参数 {name} 提供的类型注解"
                f" {hint} 无法用于注入任何依赖，请检查是否有误"
            )

        for data in self.metadatas:
            if isinstance(data, Reflect):
                self.orig_getter = cast(
                    Callable[[], Any],
                    lambda g=self.orig_getter: Reflection(cast(Callable[[], Any], g)),
                )
                break

        super().__init__(self.orig_getter, sub_getter=None, cache=False, recursive=False)

    def _unmatch_exc(self, real_type: Any) -> DependNotMatched:
        return DependNotMatched(
            f"函数 {self.func_name} 的参数 {self.arg_name} " f"与注解 {self.hint} 不匹配",
            self.func_name,
            self.arg_name,
            real_type,
            self.hint,
        )

    def _match_check(self, val: Any) -> None:
        for data in self.metadatas:
            if isinstance(data, Exclude):
                if any(isinstance(val, t) for t in data.types):
                    raise self._unmatch_exc(type(val))

        for data in self.metadatas:
            if isinstance(data, CustomLogger):
                return

        if not is_type(val, self.hint):
            raise self._unmatch_exc(type(val))

    async def fulfill(self, dep_scope: dict[Depends, Any]) -> Any:
        val = await super().fulfill(dep_scope)

        if isinstance(val, Reflection):
            inner_val = val.__obj_getter__()
            if isawaitable(inner_val):
                raise AttributeError(
                    f"异步依赖项不能通过 {Reflect.__name__} 创建反射依赖"
                )

            self._match_check(inner_val)
            return val

        self._match_check(val)
        return val


@dataclass
class Exclude:
    """数据类。`types` 指定的类别会在依赖注入时被排除

    .. code:: python

        # 假设有继承关系 A <- B, A <- C, A <- D
        # 表示 A 中不包括 B 和 C 类别的所有子类型，当然，还是会兼容 A 类型本身
        NewTypeHint = Annotated[A, Exclude(types=[B, C])]
    """

    types: Sequence[type]


@dataclass
class CustomLogger:
    """数据类。`getter` 参数会用于指定类别日志器不存在时的获取方法

    .. code:: python

        # 如果 bot 设置的 logger 是 MyLogger 类型，则成功依赖注入
        # 否则使用 getter 获取一个日志器
        NewLoggerHint = Annotated[MyLogger, CustomLogger(getter=lambda: MyLogger())]
    """

    getter: Callable[[], Any]


@dataclass
class Reflect:
    """数据类。指定不直接获取当前依赖项，而是获取对应的一个反射代理

    这适用于希望依赖会随着上下文改变，而动态变化的情况。例如动态引用会话流程中的事件对象

    .. code:: python

        # 注入一个依赖时进一步包装为反射依赖
        event_proxy = Annotated[Event, Reflect()]
        # 就像使用 event 一样使用 event_proxy
        event_proxy.attr_xxx
        event_proxy.method_xxx()

        # 不过 event_proxy 不是完美的代理
        # 因此 isinstance 类似的操作，使用 __origin__ 获取原始对象
        isinstance(event_proxy.__origin__, SomeEventType)
        # 或者是作为运行逻辑未知的函数的参数
        dont_know_what_this_do(event_proxy.__origin__)
    """


class Reflection:
    def __init__(self, getter: Callable[[], Any]) -> None:
        super().__setattr__("__obj_getter__", getter)

    @property
    def __origin__(self) -> Any:
        return self.__obj_getter__()

    def __getattr__(self, name: str) -> Any:
        getter = self.__obj_getter__
        if name == "__obj_getter__":
            return getter
        if name.startswith("_"):
            raise AttributeError(f"在反射对象上，不允许访问名称以 _ 开头的属性：{name}")

        return getattr(getter(), name)

    def __setattr__(self, name: str, value: Any) -> Any:
        getter = self.__obj_getter__
        if name == "__obj_getter__":
            return getter
        if name.startswith("_"):
            raise AttributeError(f"在反射对象上，不允许修改名称以 _ 开头的属性：{name}")

        return setattr(getter(), name, value)


def _init_auto_deps(func: Callable[P, T], allow_manual_arg: bool) -> None:
    try:
        sign = signature(func)
    except ValueError as e:
        tip = "no signature found for builtin"
        if str(e).startswith(tip) and version_info <= (3, 10):
            raise DependInitError(
                f"内建函数 {func} 在 python <= 3.10 的版本中，无法进行依赖注入"
            ) from None
        raise

    empty = Parameter.empty
    origin_f = unwrap(func, stop=lambda f: hasattr(f, "__signature__"))
    ds = deque(origin_f.__defaults__) if origin_f.__defaults__ is not None else deque()
    kwds = origin_f.__kwdefaults__ if origin_f.__kwdefaults__ is not None else {}
    nargs: list[Any] = []

    for name, param in sign.parameters.items():
        if param.kind in (Parameter.VAR_POSITIONAL, Parameter.VAR_KEYWORD):
            continue

        if param.default is not empty:
            if name in kwds:
                pass
            else:
                ds.popleft()
                nargs.append(param.default)
            continue

        if param.annotation is empty:
            continue

        try:
            dep = AutoDepends(func, name, param.annotation)
        except DependInitError:
            if allow_manual_arg:
                continue
            raise

        if dep is None:
            continue

        if param.kind is Parameter.KEYWORD_ONLY:
            kwds[name] = dep
        else:
            nargs.append(dep)

    origin_f.__defaults__ = tuple(nargs) if len(nargs) else None
    origin_f.__kwdefaults__ = kwds if len(kwds) else None  # type: ignore[assignment]


def _get_bound_args(
    func: Callable, /, *args: Any, **kwargs: Any
) -> tuple[list[Any], dict[str, Any]]:
    sign = signature(func)

    try:
        bind = sign.bind(*args, **kwargs)
    except TypeError as e:
        fname = get_obj_name(func, otype="callable")
        raise DependBindError(
            f"依赖注入匹配失败。匹配函数 {fname} 的参数时发生错误：{e}。"
            "这可能是因为传参个数不匹配，或提供了错误的类型注解"
        ) from None

    bind.apply_defaults()
    return list(bind.args), bind.kwargs


class DependsHook(Depends, BetterABC, Generic[T]):
    """依赖钩子

    包装一个依赖项，依赖满足后内部的 hook 将会执行
    """

    def __init__(
        self,
        func: Callable[P, T] | AsyncCallable[P, T],
        cache: bool = False,
        recursive: bool = False,
    ) -> None:
        super().__init__(func, cache=cache, recursive=recursive)

    @abstractmethod
    async def deps_callback(self, val: T) -> None:
        """所有依赖钩子子类必须实现该抽象方法

        :param val: 依赖项被满足后的值
        """
        raise NotImplementedError

    async def fulfill(self, dep_scope: dict[Depends, Any]) -> Any:
        val = await super().fulfill(dep_scope)
        await self.deps_callback(val)
        return val


def inject_deps(
    injectee: Callable[..., T] | AsyncCallable[..., T], manual_arg: bool = False
) -> AsyncCallable[..., T]:
    """依赖注入标记装饰器，标记当前对象需要被依赖注入

    可以标记的对象类别有：
    同步函数，异步函数，匿名函数，同步生成器函数，异步生成器函数，实例方法、类方法、静态方法

    :param injectee: 需要被注入的对象
    :param manual_arg: 当前对象标记需要依赖注入后，是否还可以给某些参数手动传参
    :return: 异步可调用对象，但保留原始参数和返回值签名
    """

    @wraps(injectee)
    async def di_wrapped(*args: P.args, **kwargs: P.kwargs) -> T:
        _args, _kwargs = _get_bound_args(injectee, *args, **kwargs)
        dep_scope: dict[Depends, Any] = {}

        for idx, _ in enumerate(_args):
            elem = _args[idx]
            if isinstance(elem, Depends):
                _args[idx] = await elem.fulfill(dep_scope)

        for idx, k in enumerate(_kwargs.keys()):
            elem = _kwargs[k]
            if isinstance(elem, Depends):
                _kwargs[k] = await elem.fulfill(dep_scope)

        ret = injectee(*_args, **_kwargs)  # type: ignore[arg-type]
        if isawaitable(ret):
            return await ret
        return ret

    if isinstance(injectee, (FunctionType, BuiltinFunctionType)):
        _init_auto_deps(injectee, manual_arg)
        return di_wrapped
    if isinstance(injectee, LambdaType):
        return di_wrapped

    raise DependInitError(
        f"{injectee} 对象不属于以下类别中的任何一种："
        "{同步函数，异步函数，匿名函数，同步生成器函数，异步生成器函数，"
        "实例方法、类方法、静态方法}，因此不能被注入依赖"
    )



from __future__ import annotations

from copy import deepcopy
from dataclasses import dataclass
from typing import Any, Callable, Coroutine

from typing_extensions import Self

from melobot.exceptions import BotException
from melobot.typ import AsyncCallable, BetterABC, LogicMode, abstractmethod
from melobot.utils import to_async

from ..adapter.event import Event


class UtilsError(BotException): ...


class Cloneable:
    def copy(self) -> Self:
        return deepcopy(self)


class Checker(BetterABC, Cloneable):
    """检查器基类"""

    def __init__(self, fail_cb: AsyncCallable[[], None] | None = None) -> None:
        super().__init__()
        self.fail_cb = fail_cb

    def __and__(self, other: Checker) -> WrappedChecker:
        if not isinstance(other, Checker):
            raise UtilsError(f"联合检查器定义时出现了非检查器对象，其值为：{other}")
        return WrappedChecker(LogicMode.AND, self, other)

    def __or__(self, other: Checker) -> WrappedChecker:
        if not isinstance(other, Checker):
            raise UtilsError(f"联合检查器定义时出现了非检查器对象，其值为：{other}")
        return WrappedChecker(LogicMode.OR, self, other)

    def __invert__(self) -> WrappedChecker:
        return WrappedChecker(LogicMode.NOT, self)

    def __xor__(self, other: Checker) -> WrappedChecker:
        if not isinstance(other, Checker):
            raise UtilsError(f"联合检查器定义时出现了非检查器对象，其值为：{other}")
        return WrappedChecker(LogicMode.XOR, self, other)

    @abstractmethod
    async def check(self, event: Event) -> bool:
        """检查器检查方法

        任何检查器应该实现此抽象方法。

        :param event: 给定的事件
        :return: 检查是否通过
        """
        raise NotImplementedError

    @staticmethod
    def new(func: Callable[[Event], bool]) -> Checker:
        return CustomChecker(func)


class CustomChecker(Checker):
    def __init__(self, func: Callable[[Event], bool]) -> None:
        super().__init__()
        self.func = func

    async def check(self, event: Event) -> bool:
        return self.func(event)


class WrappedChecker(Checker):
    """合并检查器

    在两个 :class:`Checker` 对象间使用 | & ^ ~ 运算符即可返回合并检查器。
    """

    def __init__(
        self,
        mode: LogicMode,
        checker1: Checker,
        checker2: Checker | None = None,
    ) -> None:
        """初始化一个合并检查器

        :param mode: 合并检查的逻辑模式
        :param checker1: 检查器1
        :param checker2: 检查器2
        """
        super().__init__()
        self.mode = mode
        self.c1, self.c2 = checker1, checker2

    def set_fail_cb(self, fail_cb: AsyncCallable[[], None] | None) -> None:
        self.fail_cb = fail_cb

    async def check(self, event: Event) -> bool:
        c2_check: Callable[[], Coroutine[Any, Any, bool | None]] = (
            (lambda: self.c2.check(event))  # type: ignore[union-attr]
            if self.c2 is not None
            else to_async(lambda: None)  # type: ignore[misc,arg-type]
        )
        status = await LogicMode.async_short_calc(
            self.mode, lambda: self.c1.check(event), c2_check
        )

        if not status and self.fail_cb is not None:
            await self.fail_cb()
        return status


class Matcher(BetterABC, Cloneable):
    """匹配器基类"""

    def __init__(self) -> None:
        super().__init__()

    def __and__(self, other: Matcher) -> WrappedMatcher:
        if not isinstance(other, Matcher):
            raise UtilsError(f"联合匹配器定义时出现了非匹配器对象，其值为：{other}")
        return WrappedMatcher(LogicMode.AND, self, other)

    def __or__(self, other: Matcher) -> WrappedMatcher:
        if not isinstance(other, Matcher):
            raise UtilsError(f"联合匹配器定义时出现了非匹配器对象，其值为：{other}")
        return WrappedMatcher(LogicMode.OR, self, other)

    def __invert__(self) -> WrappedMatcher:
        return WrappedMatcher(LogicMode.NOT, self)

    def __xor__(self, other: Matcher) -> WrappedMatcher:
        if not isinstance(other, Matcher):
            raise UtilsError(f"联合匹配器定义时出现了非匹配器对象，其值为：{other}")
        return WrappedMatcher(LogicMode.XOR, self, other)

    @abstractmethod
    async def match(self, text: str) -> bool:
        """匹配器匹配方法

        任何匹配器应该实现此抽象方法。

        :param text: 消息事件的文本内容
        :return: 是否匹配
        """
        raise NotImplementedError


class WrappedMatcher(Matcher):
    """合并匹配器

    在两个 :class:`Matcher` 对象间使用 | & ^ ~ 运算符即可返回合并匹配器
    """

    def __init__(
        self,
        mode: LogicMode,
        matcher1: Matcher,
        matcher2: Matcher | None = None,
    ) -> None:
        """初始化一个合并匹配器

        :param mode: 合并匹配的逻辑模式
        :param matcher1: 匹配器1
        :param matcher2: 匹配器2
        """
        super().__init__()
        self.mode = mode
        self.m1, self.m2 = matcher1, matcher2

    async def match(self, text: str) -> bool:
        m2_match: Callable[[], Coroutine[Any, Any, bool | None]] = (
            (lambda: self.m2.match(text))  # type: ignore[union-attr]
            if self.m2 is not None
            else to_async(lambda: None)  # type: ignore[misc,arg-type]
        )
        return await LogicMode.async_short_calc(
            self.mode, lambda: self.m1.match(text), m2_match
        )


@dataclass
class ParseArgs:
    """解析参数"""

    name: str
    vals: list[Any]


class Parser(BetterABC):
    """解析器基类

    解析器一般用作从消息文本中按规则批量提取参数
    """

    def __init__(self) -> None:
        super().__init__()

    @abstractmethod
    async def parse(self, text: str) -> ParseArgs | None:
        """解析方法

        任何解析器应该实现此抽象方法

        :param text: 消息文本内容
        :return: 解析结果

        """
        raise NotImplementedError



import asyncio
from typing import Any

from typing_extensions import TypeVar

from ..adapter.model import Event
from ..handle.base import EventHandler
from ..typ import HandleLevel
from ..utils import RWContext

KeyT = TypeVar("KeyT", bound=float, default=float)
ValT = TypeVar("ValT", default=Any)


class _KeyOrderDict(dict[KeyT, ValT]):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self.update(*args, **kwargs)
        self.__buf: list[tuple[KeyT, ValT]] = []

    def __setitem__(self, key: KeyT, value: ValT) -> None:
        if len(self) == 0:
            return super().__setitem__(key, value)

        if key <= next(reversed(self.items()))[0]:
            return super().__setitem__(key, value)

        cnt = 0
        for k, _ in reversed(self.items()):
            if key > k:
                cnt += 1
            else:
                break

        for _ in range(cnt):
            self.__buf.append(self.popitem())
        super().__setitem__(key, value)
        while len(self.__buf):
            super().__setitem__(*self.__buf.pop())

        return None

    def update(self, *args: Any, **kwargs: Any) -> None:
        for k, v in dict(*args, **kwargs).items():
            self[k] = v

    def setdefault(self, key: KeyT, default: ValT) -> ValT:
        if key not in self:
            self[key] = default
        return self[key]


class Dispatcher:
    def __init__(self) -> None:
        self.handlers: _KeyOrderDict[HandleLevel, set[EventHandler]] = _KeyOrderDict()
        self.broadcast_ctrl = RWContext()
        self.gc_interval = 5

    def internal_add(self, *handlers: EventHandler) -> None:
        for h in handlers:
            self.handlers.setdefault(h.flow.priority, set()).add(h)
            h.flow.on_priority_reset(
                lambda new_prior, h=h: self._reset_hook(h, new_prior)
            )

    async def add(self, *handlers: EventHandler) -> None:
        async with self.broadcast_ctrl.write():
            self.internal_add(*handlers)

    async def _remove(self, *handlers: EventHandler) -> None:
        for h in handlers:
            await h.expire()
            h_set = self.handlers[h.flow.priority]
            h_set.remove(h)
            if len(h_set) == 0:
                self.handlers.pop(h.flow.priority)

    async def expire(self, *handlers: EventHandler) -> None:
        async with self.broadcast_ctrl.write():
            await self._remove(*handlers)

    async def _reset_hook(self, handler: EventHandler, new_prior: HandleLevel) -> None:
        if handler.flow.priority == new_prior:
            return

        async with self.broadcast_ctrl.write():
            old_prior = handler.flow.priority
            if old_prior == new_prior:
                return
            h_set = self.handlers[old_prior]
            h_set.remove(handler)
            if len(h_set) == 0:
                self.handlers.pop(old_prior)
            self.handlers.setdefault(new_prior, set()).add(handler)

    async def broadcast(self, event: Event) -> None:
        async with self.broadcast_ctrl.read():
            for h_set in self.handlers.values():
                tasks = tuple(asyncio.create_task(h.handle(event)) for h in h_set)
                await asyncio.wait(tasks)
                if not event.spread:
                    break

    async def timed_gc(self) -> None:
        while True:
            await asyncio.sleep(self.gc_interval)
            async with self.broadcast_ctrl.write():
                hs = tuple(
                    h for h_set in self.handlers.values() for h in h_set if h.invalid
                )
                await self._remove(*hs)



from enum import Enum
from typing import Literal, Optional, cast

from melobot.typ import AsyncCallable

from ..adapter.event import Event, GroupMessageEvent, MessageEvent, PrivateMessageEvent
from ..adapter.segment import AtSegment
from .abc import Checker


class LevelRole(int, Enum):
    """用户权限等级枚举"""

    OWNER = 1 << 4
    SU = 1 << 3
    WHITE = 1 << 2
    NORMAL = 1 << 1
    BLACK = 1


class GroupRole(int, Enum):
    """群权限等级枚举"""

    OWNER = 1 << 2
    ADMIN = 1 << 1
    MEMBER = 1


class MsgChecker(Checker):
    """消息事件分级权限检查器

    主要分 主人、超级用户、白名单用户、普通用户、黑名单用户 五级
    """

    def __init__(
        self,
        role: LevelRole | GroupRole,
        owner: Optional[int] = None,
        super_users: Optional[list[int]] = None,
        white_users: Optional[list[int]] = None,
        black_users: Optional[list[int]] = None,
        fail_cb: Optional[AsyncCallable[[], None]] = None,
    ) -> None:
        """初始化一个消息事件分级权限检查器

        :param role: 允许的等级（>= 此等级才能通过校验）
        :param owner: 主人的 qq 号
        :param super_users: 超级用户 qq 号列表
        :param white_users: 白名单用户 qq 号列表
        :param black_users: 黑名单用户 qq 号列表
        :param fail_cb: 检查不通过的回调
        """
        super().__init__(fail_cb)
        self.check_role = role

        self.owner = owner
        self.su_list = super_users if super_users is not None else []
        self.white_list = white_users if white_users is not None else []
        self.black_list = black_users if black_users is not None else []

    def _get_level(self, event: MessageEvent) -> LevelRole | GroupRole:
        """获得事件对应的登记"""

        if isinstance(self.check_role, LevelRole):
            qid = event.user_id

            if qid in self.black_list:
                return LevelRole.BLACK
            if qid == self.owner:
                return LevelRole.OWNER
            if qid in self.su_list:
                return LevelRole.SU
            if qid in self.white_list:
                return LevelRole.WHITE
            return LevelRole.NORMAL

        if not event.is_group():
            return cast(GroupRole, GroupRole.MEMBER >> 1)
        if event.sender.is_group_owner():
            return GroupRole.OWNER
        if event.sender.is_group_admin():
            return GroupRole.ADMIN
        return GroupRole.MEMBER

    def _check(self, event: MessageEvent) -> bool:
        e_level = self._get_level(event)
        if isinstance(e_level, LevelRole):
            status = LevelRole.BLACK < e_level and e_level >= self.check_role
        else:
            status = e_level >= self.check_role
        return status

    async def check(self, event: Event) -> bool:
        if not isinstance(event, MessageEvent):
            status = False
        else:
            status = self._check(event)

        if not status and self.fail_cb is not None:
            await self.fail_cb()
        return status


class GroupMsgChecker(MsgChecker):
    """群聊消息事件分级权限检查器

    基本功能与 :class:`MsgChecker` 一致。但增加了白名单机制，
    不提供 `white_groups` 参数默认拒绝所有群聊消息事件。

    对所有私聊消息事件校验不通过。
    """

    def __init__(
        self,
        role: LevelRole | GroupRole,
        owner: Optional[int] = None,
        super_users: Optional[list[int]] = None,
        white_users: Optional[list[int]] = None,
        black_users: Optional[list[int]] = None,
        white_groups: Optional[list[int]] = None,
        fail_cb: Optional[AsyncCallable[[], None]] = None,
    ) -> None:
        """初始化一个群聊消息事件分级权限检查器

        :param role: 允许的等级（>= 此等级才能通过校验）
        :param owner: 主人的 qq 号
        :param super_users: 超级用户 qq 号列表
        :param white_users: 白名单用户 qq 号列表
        :param black_users: 黑名单用户 qq 号列表
        :param white_groups: 白名单群号列表（不在其中的群不通过校验）
        :param fail_cb: 检查不通过的回调
        """
        super().__init__(role, owner, super_users, white_users, black_users, fail_cb)
        self.white_group_list = white_groups if white_groups is not None else []

    def _check(self, event: MessageEvent) -> bool:
        if isinstance(event, PrivateMessageEvent):
            return False
        if len(self.white_group_list) == 0:
            return False
        if cast(GroupMessageEvent, event).group_id not in self.white_group_list:
            return False

        return super()._check(event)


class PrivateMsgChecker(MsgChecker):
    """私聊消息事件分级权限检查器

    基本功能与 :class:`MsgChecker` 一致。

    对所有群聊消息事件校验不通过。
    """

    def __init__(
        self,
        role: LevelRole,
        owner: Optional[int] = None,
        super_users: Optional[list[int]] = None,
        white_users: Optional[list[int]] = None,
        black_users: Optional[list[int]] = None,
        fail_cb: Optional[AsyncCallable[[], None]] = None,
    ) -> None:
        """初始化一个私聊消息事件分级权限检查器

        :param role: 允许的等级（>= 此等级才能通过校验）
        :param owner: 主人的 qq 号
        :param super_users: 超级用户 qq 号列表
        :param white_users: 白名单用户 qq 号列表
        :param black_users: 黑名单用户 qq 号列表
        :param fail_cb: 检查不通过的回调
        """
        super().__init__(role, owner, super_users, white_users, black_users, fail_cb)

    def _check(self, event: MessageEvent) -> bool:
        if not event.is_private():
            return False

        return super()._check(event)


class MsgCheckerFactory:
    """消息事件分级权限检查器的工厂

    预先存储检查依据（各等级数据），指定检查等级后，可返回一个 :class:`MsgChecker` 类的对象
    """

    def __init__(
        self,
        owner: Optional[int] = None,
        super_users: Optional[list[int]] = None,
        white_users: Optional[list[int]] = None,
        black_users: Optional[list[int]] = None,
        white_groups: Optional[list[int]] = None,
        fail_cb: Optional[AsyncCallable[[], None]] = None,
    ) -> None:
        """初始化一个消息事件分级权限检查器的工厂

        :param owner: 主人的 qq 号
        :param super_users: 超级用户 qq 号列表
        :param white_users: 白名单用户 qq 号列表
        :param black_users: 黑名单用户 qq 号列表
        :param white_groups: 白名单群号列表（不在其中的群不通过校验）
        :param fail_cb: 检查不通过的回调（这将自动附加到生成的检查器上）
        """
        self.owner = owner
        self.su_list = super_users if super_users is not None else []
        self.white_list = white_users if white_users is not None else []
        self.black_list = black_users if black_users is not None else []
        self.white_group_list = white_groups if white_groups is not None else []

        self.fail_cb = fail_cb

    def get_base(
        self,
        role: LevelRole | GroupRole,
        fail_cb: Optional[AsyncCallable[[], None]] = None,
    ) -> MsgChecker:
        """根据内部依据和给定等级，生成一个 :class:`MsgChecker` 对象

        :param role: 允许的等级（>= 此等级才能通过校验）
        :param fail_cb: 检查不通过的回调
        :return: 消息事件分级权限检查器
        """
        return MsgChecker(
            role,
            self.owner,
            self.su_list,
            self.white_list,
            self.black_list,
            self.fail_cb if fail_cb is None else fail_cb,
        )

    def get_group(
        self,
        role: LevelRole | GroupRole,
        fail_cb: Optional[AsyncCallable[[], None]] = None,
    ) -> GroupMsgChecker:
        """根据内部依据和给定等级，生成一个 :class:`GroupMsgChecker` 对象

        :param role: 允许的等级（>= 此等级才能通过校验）
        :param fail_cb: 检查不通过的回调
        :return: 群聊消息事件分级权限检查器
        """
        return GroupMsgChecker(
            role,
            self.owner,
            self.su_list,
            self.white_list,
            self.black_list,
            self.white_group_list,
            self.fail_cb if fail_cb is None else fail_cb,
        )

    def get_private(
        self,
        role: LevelRole,
        fail_cb: Optional[AsyncCallable[[], None]] = None,
    ) -> PrivateMsgChecker:
        """根据内部依据和给定等级，生成一个 :class:`PrivateMsgChecker` 对象

        :param role: 允许的等级（>= 此等级才能通过校验）
        :param fail_cb: 检查不通过的回调
        :return: 私聊消息事件分级权限检查器
        """
        return PrivateMsgChecker(
            role,
            self.owner,
            self.su_list,
            self.white_list,
            self.black_list,
            self.fail_cb if fail_cb is None else fail_cb,
        )


class AtMsgChecker(Checker):
    """艾特消息事件检查器"""

    def __init__(
        self,
        qid: int | Literal["all"] | None = None,
        fail_cb: Optional[AsyncCallable[[], None]] = None,
    ) -> None:
        """初始化一个艾特消息事件检查器

        :param qid: 被艾特的 qq 号。为空则接受所有艾特消息事件；不为空则只接受指定 qid 被艾特的艾特消息事件
        :param fail_cb: 检查不通过的回调
        """
        super().__init__(fail_cb)
        self.qid = qid

    async def check(self, event: Event) -> bool:
        if not isinstance(event, MessageEvent):
            return False

        qids = [seg.data["qq"] for seg in event.message if isinstance(seg, AtSegment)]
        if self.qid is None:
            return len(qids) > 0
        status = any(id == self.qid for id in qids)

        if not status and self.fail_cb is not None:
            await self.fail_cb()
        return status



from .abc import Checker, Matcher, ParseArgs, Parser
from .check import (
    AtMsgChecker,
    GroupMsgChecker,
    GroupRole,
    LevelRole,
    MsgChecker,
    MsgCheckerFactory,
    PrivateMsgChecker,
)
from .match import ContainMatcher, EndMatcher, FullMatcher, RegexMatcher, StartMatcher
from .parse import CmdArgFormatter, CmdParser, CmdParserFactory, FormatInfo



import asyncio
import base64
import inspect
import time
from contextlib import asynccontextmanager
from datetime import datetime
from functools import wraps
from typing import (
    Any,
    AsyncContextManager,
    AsyncGenerator,
    Awaitable,
    Callable,
    ContextManager,
    Coroutine,
    Literal,
    cast,
)

from typing_extensions import TypeVar

from .exceptions import ValidateError
from .typ import AsyncCallable, P, T


def get_obj_name(
    obj: Any,
    otype: Literal["callable", "class", "object"] | str = "object",
    default: str = "<anonymous %s>",
) -> str:
    """获取一个对象的限定名称或名称，这适用于一些类型较宽的参数。

    无法获取有效名称时，产生一个 `default % otype` 字符串

    例如某处接受一个 `Callable` 类型的参数，对于一般函数来说，使用
    `__qualname__` 或 `__name__` 可获得名称，但某些可调用对象这些值可能为 `None`
    或不存在。使用此方法可保证一定返回字符串

    .. code:: python

        def _(a: Callable) -> None:
            valid_str: str = get_obj_name(a, otype="callable")

        def _(a: type) -> None:
            valid_str: str = get_obj_name(a, otype="class")

        def _(a: Any) -> None:
            valid_str: str = get_obj_name(a, otype="type of a, only for str concat")


    :param obj: 对象
    :param otype: 预期的对象类型
    :param default: 无法获取任何有效名称时的默认字符串
    :return: 对象名称或默认字符串
    """
    if hasattr(obj, "__qualname__"):
        return cast(str, obj.__qualname__)

    if hasattr(obj, "__name__"):
        return cast(str, obj.__name__)

    return default % otype


def singleton(cls: Callable[P, T]) -> Callable[P, T]:
    """单例装饰器

    :param cls: 需要被单例化的可调用对象
    :return: 需要被单例化的可调用对象
    """
    obj_map = {}

    @wraps(cls)
    def wrapped(*args: P.args, **kwargs: P.kwargs) -> T:
        if cls not in obj_map:
            obj_map[cls] = cls(*args, **kwargs)
        return obj_map[cls]

    return wrapped


class RWContext:
    """异步读写上下文

    提供异步安全的读写上下文。在读取时可以多读，同时读写互斥。

    使用方法：

    .. code:: python

       rwc = RWContext()
       # 读时使用此控制器的安全读上下文：
       async with rwc.read():
           ...
       # 写时使用此控制器的安全写上下文：
       async with rwc.write():
           ...
    """

    def __init__(self, read_limit: int | None = None) -> None:
        """初始化异步读写上下文

        :param read_limit: 读取的数量限制，为空则不限制
        """
        self.write_semaphore = asyncio.Semaphore(1)
        self.read_semaphore = asyncio.Semaphore(read_limit) if read_limit else None
        self.read_num = 0
        self.read_num_lock = asyncio.Lock()

    @asynccontextmanager
    async def read(self) -> AsyncGenerator[None, None]:
        """上下文管理器，展开一个关于该对象的安全异步读上下文"""
        if self.read_semaphore:
            await self.read_semaphore.acquire()

        async with self.read_num_lock:
            if self.read_num == 0:
                await self.write_semaphore.acquire()
            self.read_num += 1

        try:
            yield
        finally:
            async with self.read_num_lock:
                self.read_num -= 1
                if self.read_num == 0:
                    self.write_semaphore.release()
                if self.read_semaphore:
                    self.read_semaphore.release()

    @asynccontextmanager
    async def write(self) -> AsyncGenerator[None, None]:
        """上下文管理器，展开一个关于该对象的安全异步写上下文"""
        await self.write_semaphore.acquire()
        try:
            yield
        finally:
            self.write_semaphore.release()


class SnowFlakeIdWorker:
    def __init__(self, datacenter_id: int, worker_id: int, sequence: int = 0) -> None:
        self.max_worker_id = -1 ^ (-1 << 3)
        self.max_datacenter_id = -1 ^ (-1 << 5)
        self.worker_id_shift = 12
        self.datacenter_id_shift = 12 + 3
        self.timestamp_left_shift = 12 + 3 + 5
        self.sequence_mask = -1 ^ (-1 << 12)
        self.startepoch = int(datetime(2022, 12, 11, 12, 8, 45).timestamp() * 1000)

        if worker_id > self.max_worker_id or worker_id < 0:
            raise ValueError("worker_id 值越界")
        if datacenter_id > self.max_datacenter_id or datacenter_id < 0:
            raise ValueError("datacenter_id 值越界")
        self.worker_id = worker_id
        self.datacenter_id = datacenter_id
        self.sequence = sequence
        self.last_timestamp = -1

    def _gen_timestamp(self) -> int:
        return int(time.time() * 1000)

    def get_id(self) -> int:
        timestamp = self._gen_timestamp()

        if timestamp < self.last_timestamp:
            raise ValueError(f"时钟回拨，{self.last_timestamp} 前拒绝 id 生成请求")
        if timestamp == self.last_timestamp:
            self.sequence = (self.sequence + 1) & self.sequence_mask
            if self.sequence == 0:
                timestamp = self._until_next_millis(self.last_timestamp)
        else:
            self.sequence = 0
        self.last_timestamp = timestamp
        new_id = (
            ((timestamp - self.startepoch) << self.timestamp_left_shift)
            | (self.datacenter_id << self.datacenter_id_shift)
            | (self.worker_id << self.worker_id_shift)
            | self.sequence
        )
        return new_id

    def get_b64_id(self, trim_pad: bool = True) -> str:
        id = base64.urlsafe_b64encode(
            self.get_id().to_bytes(8, byteorder="little")
        ).decode()
        if trim_pad:
            id = id.rstrip("=")
        return id

    def _until_next_millis(self, last_time: int) -> int:
        timestamp = self._gen_timestamp()
        while timestamp <= last_time:
            timestamp = self._gen_timestamp()
        return timestamp


_DEFAULT_ID_WORKER = SnowFlakeIdWorker(1, 1, 0)


def get_id() -> str:
    """从 melobot 内部 id 获取器获得一个 id 值，不保证线程安全。算法使用雪花算法

    :return: id 值
    """
    return _DEFAULT_ID_WORKER.get_b64_id()


def to_async(
    obj: Callable[P, T] | AsyncCallable[P, T] | Awaitable[T]
) -> Callable[P, Coroutine[Any, Any, T]]:
    """异步包装函数

    将一个可调用对象或可等待对象装饰为异步函数

    :param obj: 需要转换的可调用对象或可等待对象
    :return: 异步函数
    """
    if inspect.iscoroutinefunction(obj):
        return obj

    async def async_wrapped(*args: P.args, **kwargs: P.kwargs) -> T:
        if not inspect.isawaitable(obj):
            ret = obj(*args, **kwargs)
        else:
            ret = obj
        if inspect.isawaitable(ret):
            return await ret
        return ret

    if not inspect.isawaitable(obj):
        async_wrapped = wraps(obj)(async_wrapped)
    return async_wrapped


def to_coro(
    obj: Callable[P, T] | AsyncCallable[P, T] | Awaitable[T],
    *args: Any,
    **kwargs: Any,
) -> Coroutine[Any, Any, T]:
    """协程包装函数

    将一个可调用对象或可等待对象装饰为异步函数，并返回对应的协程

    :param obj: 需要包装的可调用对象或可等待对象
    :param args: 需要使用的位置参数
    :param kwargs: 需要使用的关键字参数
    :return: 协程
    """
    if inspect.iscoroutine(obj):
        return obj
    return to_async(obj)(*args, **kwargs)  # type: ignore[arg-type]


CbRetT = TypeVar("CbRetT", default=Any)
FirstCbRetT = TypeVar("FirstCbRetT", default=Any)
SecondCbRetT = TypeVar("SecondCbRetT", default=Any)
OriginRetT = TypeVar("OriginRetT", default=Any)


def if_not(
    condition: Callable[[], Any] | AsyncCallable[[], Any] | bool,
    reject: AsyncCallable[[], None],
    give_up: bool = False,
) -> Callable[[AsyncCallable[P, T]], AsyncCallable[P, T | None]]:
    """条件判断装饰器

    :param condition: 用于判断的条件（如果是可调用对象，则先求值再转为 bool 值）
    :param reject: 当条件为 `False` 时，执行的回调
    :param give_up: 在条件为 `False` 时，是否放弃执行被装饰函数
    """

    def deco_func(func: AsyncCallable[P, T]) -> AsyncCallable[P, T | None]:

        @wraps(func)
        async def wrapped_func(*args: P.args, **kwargs: P.kwargs) -> T | None:
            if not isinstance(condition, bool):
                ret = condition()
                status = bool(await ret if inspect.isawaitable(ret) else ret)
            else:
                status = condition

            if not status:
                await reject()

            if status or not give_up:
                return await func(*args, **kwargs)
            return None

        return wrapped_func

    return deco_func


def unfold_ctx(
    getter: Callable[[], ContextManager | AsyncContextManager],
) -> Callable[[AsyncCallable[P, T]], AsyncCallable[P, T]]:
    """上下文装饰器

    展开一个上下文，供被装饰函数使用。
    但注意此装饰器不支持获取上下文管理器 `yield` 的值

    :param getter: 上下文管理器获取方法
    """

    def deco_func(func: AsyncCallable[P, T]) -> AsyncCallable[P, T]:

        @wraps(func)
        async def wrapped_func(*args: P.args, **kwargs: P.kwargs) -> T:
            manager = getter()
            if isinstance(manager, ContextManager):
                with manager:
                    return await func(*args, **kwargs)
            else:
                async with manager:
                    return await func(*args, **kwargs)

        return wrapped_func

    return deco_func


def lock(
    callback: AsyncCallable[[], CbRetT] | None = None
) -> Callable[[AsyncCallable[P, OriginRetT]], AsyncCallable[P, CbRetT | OriginRetT]]:
    """锁装饰器

    本方法作为异步函数的装饰器使用，可以为被装饰函数加锁。

    在获取锁冲突时，调用 `callback` 获得一个回调并执行。回调执行完毕后直接返回。

    `callback` 参数为空，只应用 :class:`asyncio.Lock` 的锁功能。

    被装饰函数的返回值：被装饰函数被执行 -> 被装饰函数返回值；执行任何回调 -> 那个回调的返回值

    :param callback: 获取锁冲突时的回调
    """
    alock = asyncio.Lock()

    def deco_func(
        func: AsyncCallable[P, OriginRetT]
    ) -> AsyncCallable[P, CbRetT | OriginRetT]:

        @wraps(func)
        async def wrapped_func(*args: P.args, **kwargs: P.kwargs) -> CbRetT | OriginRetT:
            if callback is not None and alock.locked():
                return await async_guard(callback)
            async with alock:
                return await async_guard(func, *args, **kwargs)

        return wrapped_func

    return deco_func


def cooldown(
    busy_callback: AsyncCallable[[], FirstCbRetT] | None = None,
    cd_callback: AsyncCallable[[float], SecondCbRetT] | None = None,
    interval: float = 5,
) -> Callable[
    [AsyncCallable[P, OriginRetT]],
    AsyncCallable[P, OriginRetT | FirstCbRetT | SecondCbRetT],
]:
    """冷却装饰器

    本方法作为异步函数的装饰器使用，可以为被装饰函数添加 cd 时间。

    如果被装饰函数已有一个在运行，此时调用 `busy_callback` 生成回调并执行。回调执行完毕后直接返回。

    `busy_callback` 参数为空，则等待已运行的运行完成。随后执行下面的“冷却”处理逻辑。

    当被装饰函数没有在运行的，但冷却时间未结束：
        - `cd_callback` 不为空：使用 `cd_callback` 生成回调并执行。
        - `cd_callback` 为空，被装饰函数持续等待，直至冷却结束再执行。

    被装饰函数的返回值：被装饰函数被执行 -> 被装饰函数返回值；执行任何回调 -> 那个回调的返回值

    :param busy_callback: 已运行时的回调
    :param cd_callback: 冷却时间未结束的回调
    :param interval: 冷却时间
    """
    alock = asyncio.Lock()
    pre_finish_t = time.perf_counter() - interval - 1

    def deco_func(
        func: AsyncCallable[P, OriginRetT]
    ) -> AsyncCallable[P, OriginRetT | FirstCbRetT | SecondCbRetT]:

        @wraps(func)
        async def wrapped_func(
            *args: P.args, **kwargs: P.kwargs
        ) -> OriginRetT | FirstCbRetT | SecondCbRetT:
            nonlocal pre_finish_t
            if busy_callback is not None and alock.locked():
                return await async_guard(busy_callback)

            async with alock:
                duration = time.perf_counter() - pre_finish_t
                if duration > interval:
                    ret = await async_guard(func, *args, **kwargs)
                    pre_finish_t = time.perf_counter()
                    return ret

                remain_t = interval - duration
                if cd_callback is not None:
                    return await async_guard(cd_callback, remain_t)

                await asyncio.sleep(remain_t)
                ret = await async_guard(func, *args, **kwargs)
                pre_finish_t = time.perf_counter()
                return ret

        return wrapped_func

    return deco_func


def semaphore(
    callback: AsyncCallable[[], CbRetT] | None = None, value: int = -1
) -> Callable[[AsyncCallable[P, OriginRetT]], AsyncCallable[P, OriginRetT | CbRetT]]:
    """信号量装饰器

    本方法作为异步函数的装饰器使用，可以为被装饰函数添加信号量控制。

    在信号量无法立刻获取时，将调用 `callback` 获得回调并执行。回调执行完毕后直接返回。

    `callback` 参数为空，只应用 :class:`asyncio.Semaphore` 的信号量功能。

    被装饰函数的返回值：被装饰函数被执行 -> 被装饰函数返回值；执行任何回调 -> 那个回调的返回值

    :param callback: 信号量无法立即获取的回调
    :param value: 信号量阈值
    """
    a_semaphore = asyncio.Semaphore(value)

    def deco_func(
        func: AsyncCallable[P, OriginRetT]
    ) -> AsyncCallable[P, OriginRetT | CbRetT]:

        @wraps(func)
        async def wrapped_func(*args: P.args, **kwargs: P.kwargs) -> OriginRetT | CbRetT:
            if callback is not None and a_semaphore.locked():
                return await async_guard(callback)
            async with a_semaphore:
                return await async_guard(func, *args, **kwargs)

        return wrapped_func

    return deco_func


def timelimit(
    callback: AsyncCallable[[], CbRetT] | None = None, timeout: float = 5
) -> Callable[[AsyncCallable[P, OriginRetT]], AsyncCallable[P, OriginRetT | CbRetT]]:
    """时间限制装饰器

    本方法作为异步函数的装饰器使用，可以为被装饰函数添加超时控制。

    超时之后，调用 `callback` 获得回调并执行，同时取消原任务。

    `callback` 参数为空，如果超时，则抛出 :class:`asyncio.TimeoutError` 异常。

    被装饰函数的返回值：被装饰函数被执行 -> 被装饰函数返回值；执行任何回调 -> 那个回调的返回值

    :param callback: 超时时的回调
    :param timeout: 超时时间
    """

    def deco_func(
        func: AsyncCallable[P, OriginRetT]
    ) -> AsyncCallable[P, OriginRetT | CbRetT]:

        @wraps(func)
        async def wrapped_func(*args: P.args, **kwargs: P.kwargs) -> OriginRetT | CbRetT:
            try:
                return await asyncio.wait_for(async_guard(func, *args, **kwargs), timeout)
            except asyncio.TimeoutError:
                if callback is None:
                    raise TimeoutError("timelimit 所装饰的任务已超时") from None
                return await async_guard(callback)

        return wrapped_func

    return deco_func


def speedlimit(
    callback: AsyncCallable[[], CbRetT] | None = None,
    limit: int = 60,
    duration: int = 60,
) -> Callable[[AsyncCallable[P, OriginRetT]], AsyncCallable[P, OriginRetT | CbRetT]]:
    """流量/速率限制装饰器（使用固定窗口算法）

    本方法作为异步函数的装饰器使用，可以为被装饰函数添加流量控制：`duration` 秒内只允许 `limit` 次调用。

    超出调用速率限制后，调用 `callback` 获得回调并执行，同时取消原任务。

    `callback` 参数为空，等待直至满足速率控制要求再调用。

    被装饰函数的返回值：被装饰函数被执行 -> 被装饰函数返回值；执行任何回调 -> 那个回调的返回值。

    :param callback: 超出速率限制时的回调
    :param limit: `duration` 秒内允许调用多少次
    :param duration: 时长区间
    """
    called_num = 0
    min_start = time.perf_counter()
    if limit <= 0:
        raise ValidateError("speedlimit 装饰器的 limit 参数必须 > 0")
    if duration <= 0:
        raise ValidateError("speedlimit 装饰器的 duration 参数必须 > 0")

    def deco_func(
        func: AsyncCallable[P, OriginRetT]
    ) -> AsyncCallable[P, OriginRetT | CbRetT]:

        @wraps(func)
        async def wrapped_func(*args: P.args, **kwargs: P.kwargs) -> OriginRetT | CbRetT:
            fut = _wrapped_func(func, *args, **kwargs)
            fut = cast(asyncio.Future[CbRetT | OriginRetT | Exception], fut)
            fut_ret = await fut
            if isinstance(fut_ret, Exception):
                raise fut_ret
            return fut_ret

        return wrapped_func

    def _wrapped_func(
        func: AsyncCallable[P, T],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> asyncio.Future:
        # 分离出来定义，方便 result_set 调用形成递归。主要逻辑通过 Future 实现，有利于避免竞争问题。
        nonlocal called_num, min_start
        passed_time = time.perf_counter() - min_start
        res_fut: Any = asyncio.get_running_loop().create_future()

        if passed_time <= duration:
            if called_num < limit:
                called_num += 1
                asyncio.create_task(result_set(func, res_fut, -1, *args, **kwargs))

            elif callback is not None:
                asyncio.create_task(result_set(callback, res_fut, -1))

            else:
                asyncio.create_task(
                    result_set(func, res_fut, duration - passed_time, *args, **kwargs)
                )
        else:
            called_num, min_start = 0, time.perf_counter()
            called_num += 1
            asyncio.create_task(result_set(func, res_fut, -1, *args, **kwargs))

        return cast(asyncio.Future, res_fut)

    async def result_set(
        func: AsyncCallable[P, T],
        fut: asyncio.Future,
        delay: float,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> None:
        """
        只有依然在当前 duration 区间内，但超出调用次数限制的，需要等待。
        随后就是递归调用。delay > 0 为需要递归的分支。
        """
        nonlocal called_num
        try:
            if delay > 0:
                await asyncio.sleep(delay)
                res = await _wrapped_func(func, *args, **kwargs)
                fut.set_result(res)
                return

            res = await async_guard(func, *args, **kwargs)
            fut.set_result(res)

        except Exception as e:
            fut.set_result(e)

    return deco_func


def call_later(callback: Callable[[], None], delay: float) -> asyncio.TimerHandle:
    """同步函数延迟调度

    在指定的 `delay` 后调度一个 `callback` 执行。`callback` 应该是同步方法。

    :param callback: 同步函数
    :param delay: 多长时间后调度
    :return: :class:`asyncio.TimerHandle` 对象
    """
    return asyncio.get_running_loop().call_later(delay, callback)


def call_at(callback: Callable[[], None], timestamp: float) -> asyncio.TimerHandle:
    """同步函数指定时间调度

    在指定的时间戳调度一个 `callback` 执行。`callback` 应该是同步方法。`timestamp` <= 当前时刻回调立即执行

    :param callback: 同步函数
    :param timestamp: 在什么时刻调度
    :return: :class:`asyncio.TimerHandle` 对象
    """
    loop = asyncio.get_running_loop()
    if timestamp <= time.time_ns() / 1e9:
        return loop.call_later(0, callback)

    return loop.call_later(timestamp - time.time_ns() / 1e9, callback)


def async_later(callback: Coroutine[Any, Any, T], delay: float) -> asyncio.Future[T]:
    """异步函数延迟调度（可自主选择是否等待）

    在指定的 `delay` 后调度一个 `callback` 执行。`callback` 是协程。

    返回一个 :class:`asyncio.Future` 对象，你可以选择等待或不等待。等待 :class:`asyncio.Future` 即是等待 `callback` 的返回值。

    注意：如果 `callback` 未完成就被取消，需要捕获 :class:`asyncio.CancelledError`。

    :param callback: 异步函数（可有返回值）
    :param delay: 多长时间后调度
    :return: :class:`asyncio.Future` 对象
    """

    async def async_cb(fut: asyncio.Future) -> None:
        try:
            await asyncio.sleep(delay)
            res = await callback
            fut.set_result(res)
        except asyncio.CancelledError:
            callback.close()

    fut: asyncio.Future[T] = asyncio.get_running_loop().create_future()
    asyncio.create_task(async_cb(fut))
    return fut


def async_at(callback: Coroutine[Any, Any, T], timestamp: float) -> asyncio.Future[T]:
    """异步函数指定时间调度（可自主选择是否等待）

    在指定的时间戳调度一个 `callback` 执行。`callback` 是协程。

    返回一个 :class:`asyncio.Future` 对象，你可以选择等待或不等待。等待 :class:`asyncio.Future` 即是等待 `callback` 的返回值。

    注意：如果 `callback` 未完成就被取消，需要捕获 :class:`asyncio.CancelledError`。

    :param callback: 异步函数（可有返回值）
    :param timestamp: 在什么时刻调度
    :return: :class:`asyncio.Future` 对象
    """
    if timestamp <= time.time_ns() / 1e9:
        return async_later(callback, 0)

    return async_later(callback, timestamp - time.time_ns() / 1e9)


def async_interval(
    callback: Callable[[], Coroutine[Any, Any, None]], interval: float
) -> asyncio.Task[None]:
    """异步函数间隔调度（类似 JavaScript 的 setInterval）

    每过时间间隔执行 `callback` 一次。`callback` 是返回协程的可调用对象（异步函数或 lambda 函数等）。

    返回一个 :class:`asyncio.Task` 对象，可使用该 task 取消调度过程。

    :param callback: 异步函数
    :param interval: 调度的间隔
    :return: :class:`asyncio.Task` 对象
    """

    async def interval_cb() -> None:
        try:
            while True:
                coro = callback()
                await asyncio.sleep(interval)
                await coro
        except asyncio.CancelledError:
            coro.close()

    t = asyncio.create_task(interval_cb())
    return t


async def async_guard(func: AsyncCallable[..., T], *args: Any, **kwargs: Any) -> T:
    """在使用异步可调用对象时，提供用户友好的验证"""
    if not callable(func):
        raise ValidateError(f"{func} 不是异步可调用对象（返回 Awaitable 的可调用对象）")

    await_obj = func(*args, **kwargs)
    if inspect.isawaitable(await_obj):
        return await await_obj
    raise ValidateError(
        f"{func} 应该是异步函数，或其他异步可调用对象（返回 Awaitable 的可调用对象）。但它返回了：{await_obj}"
    )
